#ifndef GUARD_audio_h
#define GUARD_audio_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file audio.h
/// @brief Class for libao wrapping.

#include <map>
#include <string>

#include <ao/ao.h>

/// @brief Namespace for libao wrapper class.
namespace Libao {

    /// @brief Wrapper class for ao_sample_format.
    class Format
    {
        friend class Device;

    public:
        void Set(unsigned int bits, unsigned int channels, unsigned int sampleRate);

        void SetDefault()
        {
            Set(16, 2, 44100);
        }

        unsigned int GetBits() const
        {
            return format_.bits;
        }

        unsigned int GetChannels() const
        {
            return format_.channels;
        }

        unsigned int GetSampleRate() const
        {
            return format_.rate;
        }

    private:
        ao_sample_format format_;
    };

    /// @brief Wrapper class for ao_option.
    class Option
    {
        friend class Device;

    public:
        Option()
            : option_(0)
        {}

        ~Option()
        {
            Free();
        }

        void Free()
        {
            ao_free_options(option_);

            option_ = 0;
        }

        void Append(const std::map<std::string, std::string>& option);

    private:
        Option(const Option&);
        Option operator=(const Option&);

        ao_option* option_;
    };

    /// @brief Wrapper class for ao_device.
    class Device
    {
    public:
        Device()
            : device_(0)
        {}

        ~Device()
        {
            Close();
        }

        bool Open(int id, Format& format)
        {
            device_ = ao_open_live(id, &format.format_, 0);

            return Check();
        }

        bool Open(int id, Format& format, Option& option)
        {
            device_ = ao_open_live(id, &format.format_, option.option_);

            return Check();
        }

        bool Check() const
        {
            return 0 != device_;
        }

        void Close()
        {
            if(Check()) {
                
                ao_close(device_);
                device_ = 0;
            }
        }

        int Play(char* buffer, unsigned int size)
        {
            return ao_play(device_, buffer, size);
        }

    private:
        Device(const Device&);
        Device operator=(const Device&);

        ao_device* device_;
    };

    /// @brief Wrapper class for library setup and provide interface.
    class AudioDriver
    {
    public:
        /// @brief Create Audio Object.
        static AudioDriver& Instance()
        {
            static AudioDriver instance_;

            return instance_;
        }

        void SetDriver(const char* audioDriver, const std::map<std::string, std::string>& option);

        bool TestDriver(const char* const driverName, const std::map<std::string, std::string>& option); 

        bool OpenDevice(unsigned int bits, unsigned int channels, unsigned int sampleRate);

        void Close();

        bool Check() const;

        void Play(char* buffer, unsigned int size);

    private:
        AudioDriver()
            : driverId_(-1)
        {
            ao_initialize();
        }
 
        /// @brief Destructor of Audio.
        ///
        /// NOTES. After call ao_shutdown(), you cannot initialize libao again probably.
        /// This is a one of reason why this class is singleton.
        ~AudioDriver()
        {
            ao_shutdown();
        }

        AudioDriver(const AudioDriver&);
        AudioDriver& operator=(const AudioDriver&);

        int driverId_;

        Format format_;
        Option option_;
        Device device_;

        bool TestDriver(const char* const driverName); 

        void GetDriverId(const char* driver);

        bool IsDefault(const char* driver) const
        {
            return driver[0] == 'd';
        }
    };

    /// @brief Convenient class to test audio driver.
    class AudioDriverTester
    {
    public:
        AudioDriverTester(int id)
        {
            Init(id);
        }

        AudioDriverTester(int id, const std::map<std::string, std::string>& option)
        {
            option_.Append(option);

            Init(id);
        }

        bool Valid() const;

    private:
        AudioDriverTester& operator=(const AudioDriverTester&);
        AudioDriverTester(const AudioDriverTester&);

        Format format_;
        Option option_;
        Device device_;

        void Init(int id)
        {
            format_.SetDefault();

            device_.Open(id, format_, option_);
        }
    };
}

inline void Libao::AudioDriver::Close()
{
    device_.Close();
}

inline bool Libao::AudioDriver::Check() const
{
    return device_.Check();
}

inline void Libao::AudioDriver::Play(char* buffer, unsigned int size)
{
    device_.Play(buffer, size);
}

inline bool Libao::AudioDriverTester::Valid() const
{
    return device_.Check();
}

#endif
