#ifndef GUARD_connector_h
#define GUARD_connector_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file connector.h
/// @brief Class for network connector header file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fstream>

#include <ace/Reactor.h>
#include <ace/INET_Addr.h>
#include <ace/SOCK_Stream.h>
#include <ace/SOCK_Connector.h>
#include <ace/Connector.h>
#include <ace/Svc_Handler.h>
#include <ace/Reactor_Notification_Strategy.h>

#include <glibmm.h>

namespace NetUtil {

    /// @brief HTTP status codes.
    enum HTTPHeaderStatus {

        UNKNOWN_ERROR = 0,
        INFORMATIONAL = 1,
        SUCCESS = 2,
        REDIRECTION = 3,
        CLIENT_ERROR = 4,
        SERVER_ERROR = 5,

        CHAR_SUCCESS = '2',
        CHAR_REDIRECTION = '3'
    };

    extern const char* httpProtocolString;
    extern const size_t httpProtocolStringLength;
    extern const size_t packageInfoLength;

    /// @brief Convert decimal to hex.
    ///
    /// If orig is less than 9, return 0~9(ascii 48~57).
    /// If orig is over than 9, return A~F(ascii 65~70).
    inline unsigned char DecToHex(const unsigned char orig)
    {
        return orig < 10 ? orig + 48 : orig + 55;
    }

    /// @brief If orig need urlencode, return true.
    /// 
    /// According to RFC1738, check is char need to urlencode.
    /// This function does not check slash char in reserved characters.
    inline bool NeedUrlEncode(unsigned char orig)
    {
        return orig > 122 ||
            orig < 33 ||
            orig > 57 && orig < 65 ||
            orig > 90 && orig < 95 ||
            orig > 33 && orig < 39 ||
            orig == 43 ||
            orig == 44 ||
            //orig == 47 ||
            orig == 96;
    }

    inline bool HasHttpProtocolString(const ACE_TCHAR* const str)
    {
        return strncasecmp(str, httpProtocolString, httpProtocolStringLength) == 0;
    }

    void UrlEncode(const unsigned char* orig, unsigned char* dest);
};

/// @brief Class for HTTP URL parsing.
class Address
{
public:
    Address(const ACE_TCHAR* str);
    ~Address();

    /// @brief Return host name.
    const char* GetRawHostName() const
    {
        return rawHostName_;
    }

    /// @brief Return port number.
    const u_short& GetPortNumber() const
    {
        return portNumber_;
    }

    const char* GetRequest() const
    {
        return request_;
    }

    bool HasAuthorStr() const
    {
        return 0 != authorizationString_;
    }

    const char* GetAuthorStr() const
    {
        return authorizationString_;
    }

private:
    u_short portNumber_;

    char* rawHostName_;
    char* request_;

    char* fullAddress_;

    char* authorizationString_;

    void SplitAddressString(const ACE_TCHAR* str);
    void ParseAuthorizationString(const ACE_TCHAR* str);
    void ParseRawHostname(const ACE_TCHAR* str);
    void ParseRequest(const ACE_TCHAR* str);
    void ParseFullAddress(const ACE_TCHAR* str);
};

class HttpRequestStrGenerator
{
public:
    HttpRequestStrGenerator(const Address& address, const ACE_TCHAR* accept = "*/*");
    ~HttpRequestStrGenerator()
    {
        delete [] encodedAuthorizationString_;
        delete [] requestStr_;
    }

    const char* GetString() const
    {
        return requestStr_;
    }

    const size_t GetLength() const
    {
        return length_;
    }

private:
    char* requestStr_;
    size_t length_;

    ACE_Byte* encodedAuthorizationString_;

    void EncodeAuthorizationString(char* str);

    size_t GetHttpMessageSize(const Address& address, const ACE_TCHAR* accept);

    size_t HttpMessageSize(const Address& address, const ACE_TCHAR* accept)
    {
        return ACE_OS::strlen(address.GetRequest()) + ACE_OS::strlen(accept) + ACE_OS::strlen(address.GetRawHostName()) + NetUtil::packageInfoLength + 57;
    }

    size_t HttpAuthorSize()
    {
        return ACE_OS::strlen(ACE_reinterpret_cast(const char*, encodedAuthorizationString_)) + 23;
    }

    void MakeHttpGetMessage(const Address& address, char* message, const size_t messageSize, const ACE_TCHAR* accept)
    {
        ACE_OS::snprintf(message,
                         messageSize,
                         "GET %s HTTP/1.0\r\n" \
                         "User-Agent: " PACKAGE_NAME "/" VERSION "\r\n" \
                         "Host: %s:%d\r\n" \
                         "Accept: %s\r\n",
                         address.GetRequest(),
                         address.GetRawHostName(),
                         address.GetPortNumber(),
                         accept);

        if(!address.HasAuthorStr())
            ACE_OS::strncat(message, "\r\n", 2);
        else
            ACE_OS::snprintf(message + ACE_OS::strlen(message),
                             25 + ACE_OS::strlen(ACE_reinterpret_cast(const char*, encodedAuthorizationString_)),
                             "Authorization: Basic %s\r\n" \
                             "\r\n",
                             encodedAuthorizationString_);
    }
};

std::auto_ptr<Address> CreateAddress(const std::string& destUrl,
                                     bool convertCurrentEncoding = false);

/// @brief Class for get file from http server.
class FileGet : public ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH>
{
    typedef ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH> FileGetHandler;

public:
    FileGet():
        recvHeader_(false),
        redirect_(false),
        notifier_(0, this, ACE_Event_Handler::WRITE_MASK),
        generator_(0),
        outfile_(0),
        redirectURL_(0)
    {}

    ~FileGet()
    {
        delete [] redirectURL_;

        ACE_OS::fclose(outfile_);
    }

    void SetGenerator(HttpRequestStrGenerator* dest)
    {
        generator_ = dest;
    }

    /// @brief Set output file.
    void SetOutFile(const ACE_TCHAR* filename)
    {
        outfile_ = ACE_OS::fopen(filename, "w");
    }

    /// @brief Receive redirect or not.
    bool Redirect()
    {
        return redirect_;
    }

    /// @brief Return redirect url.
    const char* RedirectURL()
    {
        return redirectURL_;
    }

    virtual int open(void *p = 0);

    virtual int handle_input(ACE_HANDLE fd = ACE_INVALID_HANDLE);
    virtual int handle_output(ACE_HANDLE fd = ACE_INVALID_HANDLE);
    virtual int handle_timeout(const ACE_Time_Value& currentTime,
                               const void* act = 0);
private:
    enum {
        ITERATIONS = 5
    };

    bool recvHeader_;
    bool redirect_;

    int iterations_;
    ACE_Reactor_Notification_Strategy notifier_;

    HttpRequestStrGenerator* generator_;

    FILE* outfile_;
    char* redirectURL_;

    bool ReadStatusCode(char code)
    {
        switch(code) {
        case NetUtil::CHAR_REDIRECTION :
            redirect_ = true;

        case NetUtil::CHAR_SUCCESS :

            return true;

        default:
            // consider not exit event loop.
            this->reactor()->end_reactor_event_loop();

            return false;
        }
    }

    int ProcessRecvBuf(char* buf, ssize_t recvCnt)
    {
        if(recvHeader_) {
            ACE_OS::fwrite(buf, 1, recvCnt, outfile_);

            return 0;

        } else
            return ReadHeader(buf, recvCnt);
    }

    int ReadHeader(char* buf, ssize_t recvCnt);

    int GetRedirectionURL(char* orig, char* startPos, ssize_t limit);
    int SkipRemainHeader(char* orig, char* startPos, ssize_t limit);
};

#endif
