#ifndef GUARD_dirwnd_h
#define GUARD_dirwnd_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
//
// Orignal widget first appear in 'Air' project(http://air.kldp.net),
// and this is compact type. 
//
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file dirbrowsewin.h
/// @brief Class for directory browse window header file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if HAVE_EXT_HASH_SET && __GNUC__ >= 3
#include <ext/hash_set>
#else
#include <set>
#endif

#include <string>

#include <gdkmm.h>

#include <gtkmm/treemodel.h>
#include <gtkmm/treestore.h>
#include <gtkmm/treeview.h>
#include <gtkmm/menu.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/separator.h>
#include <gtkmm/button.h>
#include <gtkmm/box.h>
#include <gtkmm/stock.h>
#include <gtkmm/paned.h>

#include "misc.h"

/// @brief Class for directory browse window.
///
/// To check existence of destination path in treestore, 
/// need external container. If you can use hase set, try hash set to 
/// external container(not standard STL yet). OTherwise use std::set.
///
/// Directory add to tree process.
///
/// 1. Add a root directory(/) at first.
///
/// 2. Add a dummy directory under root.
///
/// 3. When expander is opened, read directories and add them.
///
/// 4. If something is appended, add a dummy directory under that.
class DirBrowseWnd
{
public:
    DirBrowseWnd();
    DirBrowseWnd(int border, const std::string& initialDirectory);
    ~DirBrowseWnd() {}

    void Show()
    {
        scrolledWindow_.show();
    }

    void Hide()
    {
        scrolledWindow_.hide();
    }

    void PackTo(Gtk::Box& box, Gtk::PackOptions option = Gtk::PACK_EXPAND_WIDGET)
    {
        box.pack_start(scrolledWindow_, option);
    }

    void AddFirstTo(Gtk::Paned& paned)
    {
        paned.add1(scrolledWindow_);
    }

    template <class T>
    void MenuPushFront(T elem)
    {
        menuPopup_.items().push_front(elem);
    }

    /// @brief Get selected directory full path.
    const std::string& SelectedDirectory() const
    {
        return selectedDirectory_;
    }

    const std::string& SelectedDirectoryFromIter();

    /// @brief Get signal handler for directory select signal.
    sigc::signal<void>& SignalSelected()
    {
        return signalSelected_;
    }

    void SetFontSize(unsigned int size)
    {
        Pango::FontDescription font = treeView_.get_style()->get_font();

        font.set_size( Pango::SCALE * size );
        Glib::RefPtr<Gtk::RcStyle> style = Gtk::RcStyle::create();
        style->set_font(font);
        treeView_.modify_style(style);
    }

private:

#if HAVE_EXT_HASH_SET && __GNUC__ >= 3
    struct StringHash
    {
        size_t operator()(const std::string& id) const
        {
            const std::string::size_type size(id.size());
            size_t h(0);
            for(std::string::size_type i(0); i < size; ++i) {

		    h = 5 * h + id[i];
            }

            return h;
        }
    };
	
    __gnu_cxx::hash_set<std::string, StringHash> pathInStore_;

    void ResetPathSet()
    {
        __gnu_cxx::hash_set<std::string, StringHash>().swap(pathInStore_);
    }

#else
    std::set<std::string> pathInStore_;

    void ResetPathSet()
    {
        std::set<std::string>().swap(pathInStore_);
    }
	
#endif

    /// @brief Class for column used in DirBrowseWnd
    class ModelColumns : public Gtk::TreeModel::ColumnRecord
    {
    public:
        ModelColumns() { add(colName_); add(colPath_); }

        // colName_ is display element, so i'll try it Glib::ustring.
        Gtk::TreeModelColumn<Glib::ustring> colName_;	///< Short path name to display.
        Gtk::TreeModelColumn<std::string> colPath_;	///< Full path name. 
    };

    ModelColumns column_;			    ///< Column used in DirBrowseWnd.	

    Gtk::ScrolledWindow scrolledWindow_;
    
    Gtk::TreeView treeView_;		    ///< Treeview to display column.

    Gtk::Menu menuPopup_;			    ///< Menu to use refresh menu.
	
    static std::string selectedDirectory_;	    ///< Selected directory.

    Glib::RefPtr<Gtk::TreeStore> treeModel_;

    sigc::signal<void> signalSelected_;	    ///< Signal handler for column select signal.

    void Init();
    void ConnectSignals();

    void OnShow();

    void OnBtnPressEvent(GdkEventButton* ev);
	
    void OnRowActivated(const Gtk::TreeModel::Path&, Gtk::TreeViewColumn*);
    void OnRowExpanded(const Gtk::TreeModel::iterator&, const Gtk::TreeModel::Path&);

    void ExpandCurrentDir();

    void OnMenuRefresh();
    void GetUnregisteredSubDirectories(const std::string& currentPath, std::vector<std::string>& dest, bool& hasSubDir);

    void InsertName(Gtk::TreeModel::Row child, const std::string& name, const std::string& fullpath);

    Gtk::TreeModel::iterator GetIterFromCursor();

    void AddDirs(std::vector<std::string>& dirNames, const Gtk::TreeModel::iterator& iter);

    Gtk::TreePath ExpandSubRows();

    /// @brief Add root directory to tree model.
    void AddRootDirectory()
    {
        Gtk::TreeModel::Row parent( *(treeModel_->append()) );
        parent[column_.colName_] = "/";
        parent[column_.colPath_] = "/";

        treeModel_->append(parent.children());
    }

    /// @brief Check event is button press and button 3.
    bool IsThirdButtonPress(const GdkEventButton* const ev)
    {
        return ev->button == 3;
    }

    /// @brief Change cursor to watch.
    ///
    /// Works only widget is realized.
    void ChangeWatchCursor()
    {
        if(scrolledWindow_.is_realized()) {
            scrolledWindow_.get_window()->set_cursor(Gdk::Cursor(Gdk::WATCH));
            while(Glib::MainContext::get_default()->iteration(false));
        }
    }

    /// @brief Change cursor to root window cursor.
    ///
    /// Works only widget is realized.
    void ChangeMainCursor()
    {
        if(scrolledWindow_.is_realized())
            scrolledWindow_.get_window()->set_cursor();
    }

    void FillRow(Gtk::TreeModel::Row& child, const std::string& name, const std::string& fullPath)
    {

        child[column_.colName_] = StringToUTF8::Instance().Convert(name);
        child[column_.colPath_] = fullPath;

        if(!access(fullPath.c_str(), X_OK | R_OK))
            treeModel_->append(child.children());
    }

    Gtk::TreeModel::iterator GetIterFrom(const Gtk::TreeModel::Path& path)
    {
        return treeModel_->get_iter(path);
    }

    Gtk::TreeModel::Row GetDummyRow(const Gtk::TreeModel::iterator& iter)
    {
        return Gtk::TreeModel::Row( *(iter->children().begin()) );
    }

    Gtk::TreeModel::Row GetNewRow(const Gtk::TreeModel::iterator& iter)
    {
        return Gtk::TreeModel::Row( *(treeModel_->append(iter->children())) );
    }

    void RemoveDummy(const Gtk::TreeModel::iterator& iter)
    {
        treeModel_->erase(iter->children().begin());
    }

    bool ValidSelectedDirectory()
    {
        return !selectedDirectory_.empty() && IsExecutable(selectedDirectory_) && IsReadable(selectedDirectory_);
    }

    void ExpandFirstRow()
    {
        treeView_.expand_row(Gtk::TreePath("0"), false);
    }

    void ScrollToPath(const Gtk::TreePath& path)
    {
        treeView_.set_cursor(path);
        treeView_.scroll_to_row(path, 0.5);
    }

    bool FindSubDirName(Glib::ustring::const_iterator& begin, Glib::ustring::const_iterator& slash, const Glib::ustring::const_iterator& end)
    {
        begin = find_if(begin, end, std::not1(IsCharSlash()));
        slash = find_if(begin, end, IsCharSlash());

        return begin != end ? true : false;
    }
};

#endif
