#ifndef GUARD_mainmanager_h
#define GUARD_mainmanager_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainmanager.h
/// @brief Class for main program control mediator header file.

#include <string>
#include <vector>
#include <map>

#include <glibmm.h>
#include <gtkmm/main.h>
#include <gtkmm/filechooserdialog.h>
#include <sigc++/sigc++.h>

#include <mad.h>

#include "fwd.h"

#include "command.h"

#include "trayicon.h"
#include "preference.h"
#include "laststate.h"
#include "enums.h"

class PlayControlAudio
{
public:
    virtual ~PlayControlAudio() {}

    void SetDriver();

    void Unpause();

private:
    virtual void DoSetDriver() = 0;

    virtual void DoUnpause() = 0;
};

inline void PlayControlAudio::SetDriver()
{
    DoSetDriver();
}

inline void PlayControlAudio::Unpause()
{
    DoUnpause();
}

class PlayControlDisplay
{
public:
    virtual ~PlayControlDisplay() {}

    virtual void UpdateWndItemRead() = 0;
    virtual void UpdateWndPlay() = 0;
    virtual void UpdateWndStop() = 0;
    virtual void UpdateWndItemRemoved() = 0;
};

class PlayerInterface
{
public:
    virtual ~PlayerInterface() {}

    virtual bool OpenDevice(unsigned int bits, unsigned int channels, unsigned int sampleRate) = 0;

    virtual bool CheckOpenDevice() = 0;

    virtual void Finished() = 0;
};

class SettingInterface
{
public:
    virtual ~SettingInterface() {}

    virtual SaguaPreference& GetPreference() = 0;

    void SavePreference();

#ifdef ENABLE_HTTP
    /// @brief Convert url to current encoding or not.
    virtual bool GetConvertUrlToCurrentEncoding() const = 0;
#endif

    virtual LastState& GetLastState() = 0;

private:
    virtual void DoSavePreference() = 0;
};

inline void SettingInterface::SavePreference()
{
    DoSavePreference();
}

class MainWndInterface
{
public:
    virtual ~MainWndInterface() {}

    virtual long int GetCurrentTime() = 0;

    virtual double GetVolumn() const = 0;

    virtual void SetVolumn(double volumn) = 0;

    virtual void About() = 0;

    virtual void Quit() = 0;

    virtual void RunPrefWnd() = 0;

    virtual bool InPlay() = 0;

    virtual void Prev() = 0;

    virtual void PlayStart() = 0;

    virtual void Stop() = 0;

    virtual void Next() = 0;

    virtual void PauseToggle() = 0;

    virtual void Seek(unsigned long time) = 0;

    virtual void SetNormalPlay() = 0;

    virtual void SetRepeatPlay() = 0;

    virtual void SetRepeatCurrentPlay(bool repeat) = 0;

    virtual void SetRandomPlay() = 0;

    virtual void TrayIconSetTooltips(const Glib::ustring& string) = 0;
};

class AddItem
{
public:
    virtual ~AddItem() {}

    virtual void FileAdd() = 0;
    virtual void DirAdd() = 0;

#ifdef ENABLE_HTTP
    virtual void UrlAdd() = 0;
#endif
};

class MainManager : public PlayControlAudio, public PlayControlDisplay, public PlayerInterface, public SettingInterface, public MainWndInterface, public AddItem
{
public:
    virtual ~MainManager() {}

    virtual void Run() = 0;

    virtual void SetChangedDriver() = 0;

    virtual bool TestDriver(const char* driver, const std::map<std::string, std::string>& options) = 0;

    virtual void Refresh() = 0;

    virtual void VolumnIncrease() = 0;

    virtual void VolumnDecrease() = 0;

    virtual void TimeUpdate() = 0;

    virtual void SetStatusbar(const Glib::ustring& text) = 0;

    virtual void SetStatusbarText(const Glib::ustring& text) = 0;

    virtual void UpdateStatusbar() = 0;

    virtual void UpdateStatusbarTimeoutClear() = 0;

    virtual void ClearStatusbar() = 0;

    virtual void TimeoutClearStatusbar() = 0;

    virtual void FileInfo() = 0;

    virtual void SetCurrentTrayIconAction() = 0;

    virtual void TrayIconFirstBtnClickAction() = 0;

    virtual void MainWndOpen() = 0;

    virtual void MainWndChangeCompactBgPixmap() = 0;

    virtual void PushList(const std::vector<std::string>& list) = 0;

    virtual void SocketReceived() = 0;

    virtual sigc::signal1<void, std::vector<std::string>&>& SignalReceived() = 0;
};

/// @brief Class for main program control mediator. 
///
/// Almost class is connected to MainManager, not to exact class.
class MainManagerImpl : public MainManager
{
public:
    /// @brief Create MainManager object. (singleton pattern)
    /// @return Pointer of created MainManager object.
    static MainManager& Instance()
    {
        static MainManagerImpl instance;

        return instance;
    }

    void Run();

    SaguaPreference& GetPreference();

    void DoSavePreference();

#ifdef ENABLE_HTTP
    bool GetConvertUrlToCurrentEncoding() const;
#endif

    void SetChangedDriver();

    bool OpenDevice(unsigned int bits, unsigned int channels, unsigned int sampleRate);

    bool CheckOpenDevice();

    bool TestDriver(const char* driver, const std::map<std::string, std::string>& options);

    void PauseToggle();

    long int GetCurrentTime();

    void Refresh();

    double GetVolumn() const;

    void SetVolumn(double volumn);

    void VolumnIncrease();

    void VolumnDecrease();

    bool InPlay();

    void Prev();

    void PlayStart();

    void Stop();

    void Next();

    void Seek(unsigned long time);

    void Finished();

    void TimeUpdate();

    void SetNormalPlay();

    void SetRepeatPlay();

    void SetRepeatCurrentPlay(bool repeat);

    void SetRandomPlay();

    void UpdateWndItemRead();

    void UpdateWndPlay();

    void UpdateWndStop();

    void UpdateWndItemRemoved();

    void FileInfo();

    void TrayIconSetTooltips(const Glib::ustring& string);

    void SetCurrentTrayIconAction();

    void TrayIconFirstBtnClickAction();

    void MainWndOpen();

    void MainWndChangeCompactBgPixmap();

    void PushList(const std::vector<std::string>& list);

    void SocketReceived();

    sigc::signal1<void, std::vector<std::string>&>& SignalReceived();

    void SetStatusbar(const Glib::ustring& text);

    void SetStatusbarText(const Glib::ustring& text);

    void UpdateStatusbar();

    void UpdateStatusbarTimeoutClear();

    void ClearStatusbar();

    void TimeoutClearStatusbar();

    void About();

    void Quit();

    void RunPrefWnd();
	
    LastState& GetLastState();

    void FileAdd();

    void DirAdd();

#ifdef ENABLE_HTTP
    void UrlAdd();
#endif

private:
    MainManagerImpl();
    ~MainManagerImpl();

    MainManagerImpl(const MainManagerImpl&);
    MainManagerImpl& operator=(const MainManagerImpl&);

    const bool isCharSetUTF8_;

    Command command_;

    SaguaTrayIcon trayIcon_;
    SaguaPreference preference_;
    LastState lastState_;
    Libao::AudioDriver& audio_;
    std::auto_ptr<SaguaAudioBuffer> audioBuffer_;
    std::auto_ptr<Audio::Mixer> mixer_;
    std::auto_ptr<PlayController> playController_;

    std::auto_ptr<Playlist> playlist_;
    std::auto_ptr<DirPlayWnd> dirPlayWnd_;
    std::auto_ptr<MainWnd> mainWnd_;

    Gtk::FileChooserDialog fileChooserDlg_;

    Glib::Dispatcher signalAudioDeviceOpenFailed_;

    sigc::signal1<void, std::vector<std::string>&> signalReceived_;

    void DoSetDriver();

    void DoUnpause();

    void SetFileChooserDlgProperty();

    void RegisterFunctionMap();

    typedef void (MainManager::* MainManagerMemFun)();
    typedef std::map<std::string, MainManagerMemFun> FunctionMap;

    FunctionMap functionMap_;
    MainManagerMemFun currentTrayIconAction_;
};

#endif
