#ifndef GUARD_misc_h
#define GUARD_misc_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file misc.h
/// @brief Miscellaneous functions.

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>

#include <map>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <cctype>

#include <glibmm.h>

#include "exception.h"

/// @brief Function object to convert char to lower case.
struct ToLower: std::unary_function<char, void>
{
    void operator()(char& p) const
    {
        p = std::tolower(p);
    }
};

/// @brief Function object to convert char to upper case.
struct ToUpper: std::unary_function<char, void>
{
    void operator()(char& p) const
    {
        p = std::toupper(p);
    }
};

/// @brief Function object to test char is white-space character.
struct IsCharSpace: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        return isspace(p);	
    }
};    

/// @brief Function object to test char is '='.
struct IsCharEqual: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        if(p == '=')
            return true;
        else
            return false;
    }
};

/// @brief Function object to test char is '/'.
struct IsCharSlash: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        if(p == '/')
            return true;
        else
            return false;    
    }
};

/// @brief Function object to test char is '.'.
struct IsCharComa: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        if(p == '.')
            return true;
        else
            return false;    
    }
};

/// @brief Test string is just white-space.
inline bool StringIsSpace(const Glib::ustring& str)
{
    return find_if(str.begin(), str.end(), std::not1(IsCharSpace())) == str.end();
}

/// @brief Test path is directory.
inline bool IsDirectory(const std::string& path)
{
    return Glib::file_test(path, Glib::FILE_TEST_IS_DIR);
}

/// @brief Test path is executable.
///
/// See also Glib::file_test
inline bool IsExecutable(const std::string& path)
{
    return (access(path.c_str(), X_OK) == 0);
}

/// @brief Test path is readable.
inline bool IsReadable(const std::string& path)
{
    return (access(path.c_str(), R_OK) == 0);
}

std::string GetExt(const char* filename);
inline std::string GetExt(const std::string& filename)
{
    return GetExt(filename.c_str());
}

/// @brief Microseconds interval sleep.
///
/// This use select function.
inline int usecsleep(const int& secs, const int& usecs)
{
    struct timeval tv;

    tv.tv_sec = secs;
    tv.tv_usec = usecs;

    return select(0, 0, 0, 0, &tv);
}

inline unsigned int StringHash(const std::string& orig)
{
    const unsigned int size = orig.size();
    unsigned int h = 0;
    for(unsigned int i = 0; i < size; ++i) {

            h = 5 * h + orig[i];
    }

    return h;
}

Glib::ustring ConvertSecondToTimeString(const unsigned long& second);

bool HasHttpProtocolString(const std::string& filePath);
bool HasFileProtocolString(const std::string& str);

bool ConfigfileToMap(const std::string& filename, std::map<std::string, std::string>& configMap, const bool& convertLowerCase = true, const bool& valueBreakAtSpace= true);

inline void ReadConfigMap(const std::map<std::string, std::string>& config, const char* str, std::string& dest)
{
    std::map<std::string, std::string>::const_iterator i = config.find(str);
    if(i != config.end())
        dest = i->second;
}

inline void ReadConfigMap(const std::map<std::string, std::string>& config, const char* str, bool& dest)
{
    std::map<std::string, std::string>::const_iterator i = config.find(str);
    if(i != config.end())
        std::strtol( (i->second).c_str(), 0, 10) == 1 ?
            dest = true : dest = false;
}

inline void ReadConfigMap(const std::map<std::string, std::string>& config, const char* str, long int& dest)
{
    std::map<std::string, std::string>::const_iterator i = config.find(str);
    if(i != config.end())
        dest = std::strtol( (i->second).c_str(), 0, 10);
}

inline bool IsComment(std::string& line)
{
    return line[0] == '#';
}

inline void StringToLower(std::string& str)
{
    for_each(str.begin(), str.end(), ToLower());
}

inline void StringToUpper(std::string& str)
{
    for_each(str.begin(), str.end(), ToUpper());
}

/// @brief Elliminate last white space character.
inline void TrimLastSpaceCharacter(std::string& dest)
{
    const std::string::reverse_iterator space( find_if(dest.rbegin(), dest.rend(), std::not1(IsCharSpace())) );
    if(space != dest.rend() && space != dest.rbegin()) {
        dest = std::string(dest.begin(), space.base());

    }
}

inline bool Unlink(const std::string& dest)
{
    if(unlink(dest.c_str()) == -1 && errno != ENOENT)
        return false;

    return true;
}

inline void IntegerToString(int orig, Glib::ustring& dest)
{
    char tmp[13];
    std::snprintf(tmp, 13, "%d", orig);
    dest = tmp;
}

class StringToUTF8
{
public:
    static StringToUTF8& Instance()
    {
        static StringToUTF8 instance;

        return instance;
    }

    Glib::ustring Convert(const Glib::ustring& orig, const std::string& fail) const
    {
        return orig.validate() ? orig :
            StringToUTF8WithFail(orig, fail, convertCharset_);
    }

    Glib::ustring Convert(const Glib::ustring& orig) const
    {
        return orig.validate() ? orig :
            StringToUTF8WithoutFail(orig, convertCharset_);
    }

    Glib::ustring Convert(const std::string& orig, const std::string& fail) const
    {
        return StringToUTF8WithFail(orig, fail, convertCharset_);
    }

    Glib::ustring Convert(const std::string& orig) const
    {
        return StringToUTF8WithoutFail(orig, convertCharset_);
    }

    Glib::ustring Convert(const char* orig, const std::string& fail) const
    {
        return StringToUTF8WithFail(orig, fail, convertCharset_);
    }

    Glib::ustring Convert(const char* orig) const
    {
        return StringToUTF8WithoutFail(orig, convertCharset_);
    }

    void SetAssumeCharset(const std::string& charset)
    {
        assumeCharset_ = charset;
    }

private:
    StringToUTF8()
        : assumeCharset_("eucKR"),
          isCharSetUTF8_(Glib::get_charset(currentCharset_)),
          convertCharset_(isCharSetUTF8_ ? assumeCharset_ : currentCharset_)
    {}

    ~StringToUTF8()
    {}

    StringToUTF8(const StringToUTF8&);
    StringToUTF8& operator=(const StringToUTF8&);

    std::string assumeCharset_;
    std::string currentCharset_;
    const bool isCharSetUTF8_;

    const std::string& convertCharset_;

    Glib::ustring StringToUTF8WithoutFail(const Glib::ustring& orig, const std::string& charset) const;

    Glib::ustring StringToUTF8WithoutFail(const std::string& orig, const std::string& charset) const;

    Glib::ustring StringToUTF8WithoutFail(const char* orig, const std::string& charset) const;

    /// @brief Convert string to UTF8 with fail string.
    /// @param orig Original string to convert.
    /// @param fail Retry string when convert is failed.
    /// @param charset From charset.
    /// @return Converted string.
    /// @sa Glib::ustring StringToUTF8
    Glib::ustring StringToUTF8WithFail(const Glib::ustring& orig, const std::string& fail, const std::string& charset) const;

    Glib::ustring StringToUTF8WithFail(const std::string& orig, const std::string& fail, const std::string& charset) const;

    Glib::ustring StringToUTF8WithFail(const char* orig, const std::string& fail, const std::string& charset) const;
};

class ProgramFile
{
public:
    static ProgramFile& Instance()
    {
        static ProgramFile instance;

        return instance;
    }

    const std::string& GetDir() const
    {
        return programDir_;
    }

    bool CheckDir() const
    {
        if(!Glib::file_test(programDir_, Glib::FILE_TEST_EXISTS)) {

            return MakeDir();

        } else if(!Glib::file_test(programDir_, Glib::FILE_TEST_IS_DIR))
            return false;
	    
        return true;
    }

    bool MakeDir() const
    {
        return 0 == mkdir(programDir_.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
    }

    const char* GetConfig() const
    {
        return configFile_;
    }

    const char* GetLastState() const
    {
        return lastStateFile_;
    }

    const char* GetMainList() const
    {
        return mainListFile_;
    }

private:
    ProgramFile()
        : programDir_(Glib::get_home_dir() + "/.sagua")
    {
        configFile_ = new char[programDir_.size() + 8];
        std::snprintf(configFile_, programDir_.size() + 8, "%s/config", programDir_.c_str());

        lastStateFile_ = new char[programDir_.size() + 11];
        std::snprintf(lastStateFile_, programDir_.size() + 11, "%s/laststate", programDir_.c_str());

        mainListFile_ = new char[programDir_.size() + 9];
        std::snprintf(mainListFile_, programDir_.size() + 9, "%s/dir.m3u", programDir_.c_str());
    }

    ~ProgramFile()
    {
        delete [] mainListFile_;
        delete [] lastStateFile_;
        delete [] configFile_;
    }

    ProgramFile(const ProgramFile&);
    ProgramFile& operator=(const ProgramFile&);

    const std::string programDir_;

    char* configFile_;
    char* lastStateFile_;
    char* mainListFile_;
};

/// @brief Convenient RAII class for open, close function.
class OpenHandler
{
public:
    /// @brief Constructor.
    /// @param path File path.
    /// @param flags Open flags option.
    /// @throw ErrnoException File open error.
    OpenHandler(const char* path, int flags) throw(Exception::ErrnoException)
        : fd_(open(path, flags))
    {
        if(-1 == fd_)
            throw Exception::ErrnoException(errno);
    }

    ~OpenHandler()
    {
        close(fd_);
    }

    /// @brief Get file descriptor.
    int Get()
    {
        return fd_;
    }

private:
    OpenHandler(const OpenHandler&);
    OpenHandler& operator=(const OpenHandler&);

    int fd_;
};

/// @brief Convenient RAII class for mmap, munmap function.
class MmapHandler
{
public:
    /// @brief Constructor.
    /// @param path File path.
    /// @param prot mmap protection option.
    /// @param flags mmap flags option.
    /// @throw ErrnoException Error on mmap.
    /// @throw FileException File is not regular file.
    MmapHandler(const char* path, int prot, int flags) throw(Exception::SaguaException)
        : file_(path, O_RDONLY),
          fdm_(MAP_FAILED)
    {
        if(-1 == fstat(file_.Get(), &stat_))
            throw Exception::ErrnoException(errno);

        if(!S_ISREG(stat_.st_mode))
            throw Exception::FileException("File is not regular file");

        fdm_ = mmap(0, stat_.st_size, prot, flags, file_.Get(), 0);
        if(MAP_FAILED == fdm_)
            throw Exception::ErrnoException(errno);
    }

    ~MmapHandler()
    {
        munmap(fdm_, stat_.st_size);
    }

    /// @brief Get mapped region.
    void* Get()
    {
        return fdm_;
    }

    /// @brief Get mapped size.
    int GetSize() const
    {
        return stat_.st_size;
    }

private:
    MmapHandler(const MmapHandler&);
    MmapHandler& operator=(const MmapHandler&);

    OpenHandler file_;
    struct stat stat_;

    void* fdm_;
};

/// @brief Convenient RAII class for fopen, fclose function.
class FileHandler
{
public:
    /// @brief Constructor.
    /// @param path File path.
    /// @param mode Open mode.
    /// @throw ErrnoException File open error.
    FileHandler(const char* path, const char* mode) throw(Exception::ErrnoException)
        : file_(fopen(path, mode))
    {
        if(0 == file_)
            throw Exception::ErrnoException(errno);
    }

    ~FileHandler()
    {
        fclose(file_);
    }

    /// @brief Get file pointer.
    FILE* Get()
    {
        return file_;
    }

private:
    FileHandler(const FileHandler&);
    FileHandler& operator=(const FileHandler&);

    FILE* file_;
};

#endif
