#ifndef GUARD_player_h
#define GUARD_player_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player.h
/// @brief Abstract class for player and factory class to produce player object header file.

#include <map>
#include <string>

#include "fixedallocator.h"
#include "exception.h"
#include "misc.h"

enum DecodeState {
    UNKNOWN = 0,
    PLAY = -1,
    STOP = -2,
    SEEK = -3,
    SEEK_REWIND = -4
};

/// @brief Abstract class for player.
class Player 
{
public:
    Player() {}
    virtual ~Player() {}

    void Start();
    void Stop();
    void Seek(unsigned long time);

private:
    virtual void DoStart() = 0;
    virtual void DoStop() = 0;
    virtual void DoSeek(unsigned long time) = 0;
};

inline void Player::Start()
{
    DoStart();
}

inline void Player::Stop()
{
    DoStop();
}

inline void Player::Seek(unsigned long time)
{
    DoSeek(time);
}

/// @brief Factory class to produce player object.
template <class Buffer>
class PlayerFactory : public SmallObject
{
public:
    PlayerFactory(Buffer& buffer)
        : buffer_(buffer)
    {}

    /// @brief Register id to callback map.
    /// @param id Identification string 
    /// @param creator callback to produce object
    /// @return Success of register.
    bool Register(const char* const id, std::auto_ptr<Player> (*creator)(Buffer&, const std::string&))
    {
        return callbacks_.insert( callbackValueType(id, creator) ).second;
    }

    /// @brief Unregister id from callback map.
    /// @param id Identification string.
    /// @return Success of unregister.
    bool Unregister(const std::string& id)
    {
        return callbacks_.erase(id) == 1;
    }

    /// @brief Check factory has filename's type.
    /// @param filename Filename to check.
    /// @return Return true if factory has filename's type.
    bool Exist(const std::string& filename) const
    {
        callbackIter i = callbacks_.find(GetExt(filename));

        if(i != callbacks_.end())
            return true;
        else
            return false;	
    }

    /// @brief Create player object.
    /// @param filename Filename to play.
    /// @return Player object.
    std::auto_ptr<Player> Create(const std::string& filename)
    {
        callbackIter i = callbacks_.find(GetExt(filename));

        if(i != callbacks_.end()) {

            try {

                return (i->second)(buffer_, filename);

            } catch(const Exception::SaguaException& ex) {

                return std::auto_ptr<Player>();
            }

        } else
            return std::auto_ptr<Player>();	
    }

private:
    typedef std::map<std::string, std::auto_ptr<Player> (*)(Buffer&, const std::string&)> callback;
    typedef typename std::map<std::string, std::auto_ptr<Player> (*)(Buffer&, const std::string&)>::value_type callbackValueType;
    typedef typename std::map<std::string, std::auto_ptr<Player> (*)(Buffer&, const std::string&)>::const_iterator callbackIter;

    callback callbacks_;	    ///< callback map
    Buffer& buffer_;
};

#endif
