#ifndef GUARD_player_ogg_h
#define GUARD_player_ogg_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_ogg.h
/// @brief Class for play ogg header file.

#include <vorbis/vorbisfile.h>

#include <glibmm.h>

#include "player.h"

#include "fwd.h"

#include "exception.h"

/// @brief Namespace for class that use ogg vorbis.
namespace VORBIS {

    /// @brief Wrapper class for OggVorbis_File.
    class OVFile
    {
    public:
        /// @brief Open file and ov_open that file.
        /// @throw ErrnoException Exception on open file.
        /// @throw OggException Exception on ov_open.
        OVFile(const char* filename)
        {
            FILE* inFile(fopen(filename, "rb"));
            if(0 == inFile)
                throw Exception::ErrnoException(errno);

            // OggVorbis_File will take ownership of file pointer.
            // So if success on ov_open, file pointer will be closed at
            // ov_clear.
            if(ov_open(inFile, &vf_, 0, 0) < 0) {

                fclose(inFile);

                throw Exception::OggException("ov_open failed");
            }
        }

        ~OVFile()
        {
            ov_clear(&vf_);
        }

        int GetChannels() const
        {
            return vf_.vi->channels;
        }

        long GetSampleRate() const
        {
            return vf_.vi->rate;
        }

        long Read(float*** buffer, int bufSize, int* bitstream)
        {
            return ov_read_float(&vf_, buffer, bufSize, bitstream);
        }

        int Seek(double pos)
        {
            return ov_time_seek(&vf_, pos);
        }

        double GetTotalTime()
        {
            return ov_time_total(&vf_, -1);
        }

        int GetCommentsNum() const
        {
            return vf_.vc->comments;
        }

        const char* GetComments(int index) const
        {
            return vf_.vc->user_comments[index];
        }

    private:
        OVFile(const OVFile&);
        OVFile operator=(const OVFile&);

        OggVorbis_File vf_;
    };

    /// @brief Class for play ogg.
    template <class Buffer>
    class OGGPlayer : public Player
    {
    public:
        OGGPlayer(Buffer& buffer, const std::string& filename);
        ~OGGPlayer()
        {}
	    
    private:
        OGGPlayer& operator=(const OGGPlayer& oggPlayer);
        OGGPlayer(const OGGPlayer& oggPlayer);
	    
        Glib::Dispatcher signalFinish_;	    ///< Dispatcher for finish signal. 

        Glib::Mutex vfMutex_;		    ///< Mutex to protect vorbisFile_. 
	    
        OVFile ovFile_;
	    
        DecodeState decodeState_;		    ///< DecodeState to check status.

        const int buflen_;			    ///< Decoded buffer size.

        Buffer& buffer_;
        PlayerInterface& playerInterface_;


        void DoStart();
        void DoStop();
        void DoSeek(unsigned long time);
    };

}

template <class Buffer>
std::auto_ptr<Player> CreateOGGPlayer(Buffer& buffer, const std::string& filename);

#endif
