#ifndef GUARD_tag_mp3_h
#define GUARD_tag_mp3_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_mp3.h
/// @brief Class for process mp3 tag header file.

#include <id3/tag.h>

#include <mad.h>

#include "tag.h"
#include "misc.h"
#include "exception.h"

/// @brief Class for process mp3 tag.
class MP3Tag : public Tag {
public:
    MP3Tag(const std::string& filename);

    ~MP3Tag()
    {}

private:
    ID3_Tag id3Tag_;

    virtual void ProcessYear();
    virtual void ProcessTrack();
    virtual void ProcessAlbum();
    virtual void ProcessGenre();
    virtual void ProcessArtist();
    virtual void ProcessComment();

    void ProcessTextField(Glib::ustring& dest, const ID3_FrameID& frameID);
    void ProcessTextFieldUTF8(Glib::ustring& dest, const ID3_FrameID& frameID);

    void ConvertGenreNumber();
};

std::auto_ptr<Tag> CreateId3Tag(const std::string& filename);

/// @brief Class to calculate time for MP3 file.
class MP3Time
{
public:
    /// @throw ErrnoException Error on mmap.
    /// @throw FileException Error on mmap.
    MP3Time(const char* filename)
        : fdm_(filename, PROT_READ, MAP_SHARED),
          duration_(mad_timer_zero),
          time_(0)
    {
        Calc();
    }

    ~MP3Time()
    {}

    const unsigned long& GetTime() const
    {
        return time_;
    }

private:
    MP3Time(const MP3Time&);
    MP3Time& operator=(const MP3Time&);

    enum FileType {
        CBR = 0,		///< Constant bitrate file.
        VBR = -1,		///< Variable bitrate file.
        XING_VBR = -2		///< Variable bitrate file with XING header.
    };

    MmapHandler fdm_;

    mad_timer_t duration_;
    struct mad_stream stream_;
    struct mad_header header_;
    struct mad_frame frame_;

    unsigned long time_;

    bool HasXingHeader(struct mad_bitptr* ptr, unsigned int bitlen)
    {
        return bitlen > 64 && mad_bit_read(ptr, 32) ==
                (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g');
    }

    bool HasXingFramesFlag(long flags)
    {
        return flags & 0x00000001L;
    }

    bool GetXingFrames(struct mad_bitptr* ptr, unsigned int bitlen, unsigned long& xingFrames);

    void InitMad()
    {
        mad_stream_init(&stream_);
        mad_header_init(&header_);
        mad_frame_init(&frame_);
        mad_stream_buffer(&stream_, static_cast<unsigned char*>(fdm_.Get()), fdm_.GetSize());
    }

    void CleanMad()
    {
        mad_frame_finish(&frame_);
        mad_header_finish(&header_);
        mad_stream_finish(&stream_);
    }

    bool DetectXingHeader(unsigned long& xingFrames);

    void Calc();

    void CalcCBR()
    {
        // this is not accuracy, but fast and can express almost exact time.
        time_ = static_cast<unsigned long>( static_cast<double>(fdm_.GetSize()) * 8.0 / static_cast<double>(header_.bitrate) );
    }

    void CalcVBR()
    {
        duration_ = mad_timer_abs(duration_);
        time_ = duration_.seconds;
    }

    void CalcXingVBR(unsigned long xingFrames)
    {
        mad_timer_multiply(&header_.duration, xingFrames);
        duration_ = header_.duration;
        duration_ = mad_timer_abs(duration_);
        time_ = duration_.seconds;
    }

    void DetectFileType(FileType& fileType, unsigned long& xingFrames);
};

#endif
