/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file aacbuffer.cc
/// @brief Class for aac buffer.

#include <cstring>
#include <memory>

#include "../include/aacbuffer.h"
#include "../include/scopeguard.h"

/// @brief Determine type.
bool FAAD::IsMP4File(FILE* const inFile)
{
    unsigned char header[8] = { 0, };
    std::fread(header, 1, 8, inFile);
    std::rewind(inFile);

    return header[4] == 'f' && header[5] == 't' && header[6] == 'y' && header[7] == 'p';
}

/// @brief Set mp4_callback_t.
std::auto_ptr<mp4ff_callback_t> FAAD::SetMP4ffCallback(FILE* inFile)
{
    std::auto_ptr<mp4ff_callback_t> mp4Cb(new mp4ff_callback_t);
    mp4Cb->read = MP4ReadCallback;
    mp4Cb->seek = MP4SeekCallback;
    mp4Cb->user_data = inFile;

    return mp4Cb;
}

int FAAD::GetAACTrack(mp4ff_t* const inFile)
{
    const int totalTracks = mp4ff_total_tracks(inFile);
    int rc = 0;
    unsigned char* buffer = 0;
    unsigned int bufferSize = 0;
    mp4AudioSpecificConfig mp4ASC;

    for(int i = 0; i < totalTracks; ++i) {
        mp4ff_get_decoder_config(inFile, i, &buffer, &bufferSize);

        if(0 != buffer) {

            rc = AudioSpecificConfig(buffer, bufferSize, &mp4ASC);
            free(buffer);
            buffer = 0;

            if(rc >= 0)
                return i;
        }
    }

    return -1;
}

int FAAD::AACBuffer::adtsSampleRate_[] = {96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 7350, 0, 0, 0};

bool FAAD::AACBuffer::InitBuffer(const size_t& size)
{
    readSize_ = size;

    buffer_ = static_cast<unsigned char*>(std::malloc(size));
    if(0 == buffer_)
        return false;
    std::memset(buffer_, 0, size);

    const size_t bread = std::fread(buffer_, 1, size, inFile_);
    bytesIntoBuffer_ = bread;
    bytesConsumed_ = 0;
    fileOffset_ = 0;

    if(bread != size)
        atEof_ = true;

    return true;
}

bool FAAD::AACBuffer::InitDecoder(faacDecHandle& decoder, unsigned long& sampleRate, unsigned char& channels)
{
    long int bread = faacDecInit(decoder, buffer_, bytesIntoBuffer_, &sampleRate, &channels);

    if(bread < 0)
        return false;

    AdvanceBuffer(bread);
    FillBuffer();

    return true;
}

void FAAD::AACBuffer::FillBuffer()
{
    if(bytesConsumed_ <= 0)
        return;

    if(bytesIntoBuffer_)
        memmove(static_cast<void*>(buffer_), static_cast<void*>(buffer_ + bytesConsumed_), bytesIntoBuffer_ * sizeof(unsigned char));

    if(!atEof_) {
        const int bread = std::fread(static_cast<void*>(buffer_ + bytesIntoBuffer_), 1, bytesConsumed_, inFile_);

        if(bread != bytesConsumed_)
            atEof_ = true;

        bytesIntoBuffer_ += bread;	
    }

    bytesConsumed_ = 0;

    if(bytesIntoBuffer_ > 3) {
        if(memcmp(buffer_, "TAG", 3) == 0)
            bytesIntoBuffer_ = 0;
    }

    if(bytesIntoBuffer_ > 11) {
        if(memcmp(buffer_, "LYRICSBEGIN", 11) == 0)
            bytesIntoBuffer_ = 0;
    }

    if(bytesIntoBuffer_ > 8) {
        if(memcmp(buffer_, "APETAGEX", 8) == 0)
            bytesIntoBuffer_ = 0;
    }
}

float FAAD::AACBuffer::GetLength()
{
    int tagSize = 0;
    if(memcmp(buffer_, "ID3", 3) != 0) {

        tagSize = (buffer_[6] << 21) | (buffer_[7] << 14) | (buffer_[8] << 7) |
            (buffer_[9] << 0);
        tagSize += 10;

        AdvanceBuffer(tagSize);
        FillBuffer();
    }

    float length = 0.0;

    if(buffer_[0] == 0xFF && (buffer_[1] & 0xF6) == 0xF0) {

        int frames = 0;
        int frameLength = 0;
        int sampleRate = 0;
        float framesPerSec = 0.0f;
        for(;; ++frames) {

            FillBuffer();
            if(bytesIntoBuffer_ > 7) {
                if(!(buffer_[0] == 0xFF && (buffer_[1] & 0xF6) == 0XF0))
                    break;

                if(frames == 0)
                    sampleRate = adtsSampleRate_[(buffer_[2]&0x3c)>>2];

                frameLength = ((static_cast<unsigned int>(buffer_[3]) & 0x3) << 11) | (static_cast<unsigned int>(buffer_[4]) << 3) | buffer_[5] >> 5;

                if(frameLength > bytesIntoBuffer_)
                    break;

                AdvanceBuffer(frameLength);

            } else
                break;
        }

        framesPerSec = sampleRate / 1024.0f;
        if(framesPerSec > 0.0f)
            length = frames / framesPerSec;
        
        std::fseek(inFile_, tagSize, SEEK_SET);
        const size_t bread = std::fread(buffer_, 1, readSize_, inFile_);
        if(bread != readSize_)
            atEof_ = true;
        else
            atEof_ = false;

        bytesIntoBuffer_ = bread;
        bytesConsumed_ = 0;
        fileOffset_ = tagSize; 

    } else if(std::memcmp(buffer_, "ADIF", 4) == 0) {
        const int skipSize = (buffer_[4] & 0x80) ? 9 : 0;
        const int bitRate =
            (static_cast<unsigned int>(buffer_[4 + skipSize] & 0x0F)<<19) |
            (static_cast<unsigned int>(buffer_[5 + skipSize])<<11) |
            (static_cast<unsigned int>(buffer_[6 + skipSize])<<3) |
            (static_cast<unsigned int>(buffer_[7 + skipSize]) & 0xE0);

        length = fileSize_;
        if(length != 0.0f && bitRate != 0)
            length = length * 8.0f / bitRate + 0.5f;
    }

    return length;
}
