/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file audio.cc
/// @brief Class for libao wrapping.

#include "../include/audio.h"
#include "../include/trace.h"

void Libao::Format::Set(unsigned int bits, unsigned int channels, unsigned int sampleRate)
{
    format_.bits = bits;
    format_.channels = channels;
    format_.rate = sampleRate;

#ifdef WORDS_BIGENDIAN
    format_.byte_format = AO_FMT_BIG;
#else
    format_.byte_format = AO_FMT_LITTLE;
#endif
}

void Libao::Option::Append(const std::map<std::string, std::string>& option)
{
    std::map<std::string, std::string>::const_iterator iter(option.begin());

    while(iter != option.end()) {

        ao_append_option(&option_, iter->first.c_str(), iter->second.c_str());

        ++iter;
    }
}

/// @brief Get driver id and option from preference, prepare to that.
///
/// If audio driver id is "default", try to set "esd" first.
void Libao::AudioDriver::SetDriver(const char* audioDriver, const std::map<std::string, std::string>& option)
{
    DEBUGTRACE("Libao::AudioDriver::SetDriverFromPreference");

    device_.Close();

    option_.Free();

    GetDriverId(audioDriver);

    if(driverId_ == -1) {

        DEBUGTRACE_MSG("Got error, select default driver!");

        driverId_ = ao_default_driver_id();

    } else {

        DEBUGTRACE_MSG("Selected not default driver : ");
        DEBUGTRACE_MSG(audioDriver);

        option_.Append(option);
    }
}

/// @brief Test to open device from driver name.
///
/// Even driver name is "default", does not try to "esd" first.
/// @param driverName Driver name to test.
/// @return Sucess of test.
bool Libao::AudioDriver::TestDriver(const char* const driverName)
{
    DEBUGTRACE("Libao::AudioDriver::TestDriver without option");

    DEBUGTRACE_MSG("Selected driver : ");
    DEBUGTRACE_MSG(driverName);

    const int driverId(ao_driver_id(driverName));

    if(driverId == -1) {

        DEBUGTRACE_MSG("Cannot find such driver id");

        return false;
    }

    AudioDriverTester tester(driverId);
    
    return tester.Valid();
}

/// @brief Test to open device from driver name.
///
/// If audio driver id is "default", try to set "esd" first.
/// @param driverName Driver name to test.
/// @param option Driver option.
/// @return Sucess of test.
bool Libao::AudioDriver::TestDriver(const char* const driverName, const std::map<std::string, std::string>& option)
{
    DEBUGTRACE("Libao::AudioDriver::TestDriver");

    DEBUGTRACE_MSG("Selected driver : ");
    DEBUGTRACE_MSG(driverName);

    if(IsDefault(driverName)) {

        if(TestDriver("esd"))
            return true;

        AudioDriverTester tester(ao_default_driver_id());

        return tester.Valid();
    }

    const int driverId(ao_driver_id(driverName));

    if(driverId == -1) {

        DEBUGTRACE_MSG("Cannot find such driver id");

        return false;
    }

    AudioDriverTester tester(driverId, option);

    return tester.Valid();
}

/// @brief Open device and start play buffer thread.
///
/// If bits over 16 or channels more than 2, return false.
/// @param bits Bits per sample.
/// @param channels Number of channels.
/// @param sampleRate Sample rate per sample.
/// @return Success of open.
/// @todo Remove bits and channels check to generalize.
bool Libao::AudioDriver::OpenDevice(unsigned int bits, unsigned int channels, unsigned int sampleRate)
{
    if(bits != 16 || channels > 2)
        // Can't control if that isn't 16 bits or more than 2 channels Now.
        return false;

    format_.Set(bits, channels, sampleRate);

    return device_.Open(driverId_, format_, option_);
}

void Libao::AudioDriver::GetDriverId(const char* driver)
{
    DEBUGTRACE("Libao::AudioDriver::GetDriverId");

    if(IsDefault(driver)) {

        if(TestDriver("esd")) {

            DEBUGTRACE_MSG("Detect esd on default driver");

            driverId_ = ao_driver_id("esd");

            return;
        }

        DEBUGTRACE_MSG("Driver is default");

        driverId_ = ao_default_driver_id();

    } else
        driverId_ = ao_driver_id(driver);
}
