/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainwin.cc
/// @brief Class for main window source file.

#include <iostream>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <map>

#include <gtkmm/main.h>
#include <gtkmm/scrollbar.h>
#include <gtkmm/messagedialog.h>

#include <libintl.h>

#include "../include/mainwin.h"
#include "../include/mainmanager.h"
#include "../include/preference.h"
#include "../include/dirplaywin.h"

#include "../include/playlist.h"
#include "../include/dirplaywin.h"
#include "../include/gtkmmutil.h"
#include "../include/misc.h"
#include "../include/trace.h"

MainWndImpl::MainWndImpl(MainWndInterface& mainWndInterface, SettingInterface& setting, PlaylistDisplay& playlist, AddItem& addItem, DirPlayWnd& dirPlayWnd)
    : repeatPlayMenu_(gettext("R_epeat play"), true),
      randomPlayMenu_(gettext("_Random play"), true),
      repeatCurrentPlayMenu_(gettext("Repeat _Current play"), true),
      listDisplayMenu_(gettext("_List"), true),
      browserDisplayMenu_(gettext("_Directory"), true),
      statusbarDisplayMenu_(gettext("_Statusbar"), true),
      titleDisplayAlignment_(0.5, 0.5, 0.85, 1.0),
      seekScaleAlignment_(0.5, 0.5, 0.85, 1.0),
      volumnAlignment_(0.5, 0.5, 1.0, 0.62),
      seekScale_(0.0, 0.0, 1.0),
      volumnScale_(0.0, 100.1, 1.0),
      playImgBtn_( PIXMAPSDIR "play.png" , PIXMAPSDIR "play_o.png", PIXMAPSDIR "play_p.png"),
      pauseImgBtn_( PIXMAPSDIR "pause.png" , PIXMAPSDIR "pause_o.png", PIXMAPSDIR "pause_p.png"),
      stopImgBtn_( PIXMAPSDIR "stop.png" , PIXMAPSDIR "stop_o.png", PIXMAPSDIR "stop_p.png"),
      prevImgBtn_( PIXMAPSDIR "prev.png" , PIXMAPSDIR "prev_o.png", PIXMAPSDIR "prev_p.png"),
      nextImgBtn_( PIXMAPSDIR "next.png" , PIXMAPSDIR "next_o.png", PIXMAPSDIR "next_p.png"),
      compactPlayImgBtn_( PIXMAPSDIR "c_play.png" , PIXMAPSDIR "c_play_o.png", PIXMAPSDIR "c_play_p.png"),
      compactPauseImgBtn_( PIXMAPSDIR "c_pause.png" , PIXMAPSDIR "c_pause_o.png", PIXMAPSDIR "c_pause_p.png"),
      compactStopImgBtn_( PIXMAPSDIR "c_stop.png" , PIXMAPSDIR "c_stop_o.png", PIXMAPSDIR "c_stop_p.png"),
      compactPrevImgBtn_( PIXMAPSDIR "c_prev.png" , PIXMAPSDIR "c_prev_o.png", PIXMAPSDIR "c_prev_p.png"),
      compactNextImgBtn_( PIXMAPSDIR "c_next.png" , PIXMAPSDIR "c_next_o.png", PIXMAPSDIR "c_next_p.png"),
      leftArrow_(Gtk::ARROW_LEFT, Gtk::SHADOW_ETCHED_IN),
      rightArrow_(Gtk::ARROW_RIGHT, Gtk::SHADOW_ETCHED_IN),
      upArrow_(Gtk::ARROW_UP, Gtk::SHADOW_ETCHED_IN),
      downArrow_(Gtk::ARROW_DOWN, Gtk::SHADOW_ETCHED_IN),
      defaultWidth_(300),
      defaultHeight_(330),
      defaultCompactWidth_(345),
      defaultCompactHeight_(82),
      defaultFoldHeight_(143),
      defaultTitleWidth_(defaultWidth_ - 104),
      prevWidth_(300),
      prevHeight_(330),
      maxScreenWidth_(-1),
      maxScreenHeight_(-1),
      compactMode_(false),
      skipTimeUpdate_(false),
      mainWndDefaultStyle_(get_modifier_style()),
      titleDisplayEventBoxDefaultStyle_(titleDisplayEventBox_.get_modifier_style()),
      playlist_(playlist),
      dirPlayWnd_(dirPlayWnd),
      clearStatusbarTimeout_(0),
      mainWndInterface_(mainWndInterface),
      setting_(setting)
{
    set_title("Sagua");
    set_resizable(true);
    set_size_request(defaultWidth_, defaultHeight_);
    set_default_icon_from_file( PIXMAPSDIR "sagua.png" );

    GetScreenSize();

    Packing(addItem);

    InitialDisplay();

    ConnectSignals();
}

Gtk::Window& MainWndImpl::GetWidget()
{
    return *this;
}

/// @brief Time display update.
///
/// Get current time from player and display it.
/// This function called by player when player's timer change in 
/// second domain.
void MainWndImpl::TimeUpdate()
{
    if(skipTimeUpdate_)
        return;

    long int currentTime(mainWndInterface_.GetCurrentTime());

    timeLabel_.set_label(ConvertSecondToTimeString(currentTime));
    seekScale_.set_value(currentTime);
}

/// @brief Callback function used when window is received delete 
/// event(window close in WM or else ...).
bool MainWndImpl::on_delete_event(GdkEventAny* ev)
{
    OnMenuQuit();

    return false;
}

/// @brief Callback function used when window is received map event.
bool MainWndImpl::on_map_event(GdkEventAny* event)
{
    Gtk::Window::on_map_event(event);

    SetSkipTimeUpdate(false);
    TimeUpdate();

    return false;
}

/// @brief Callback function used when window is received unmap event(hide, iconify ...).
bool MainWndImpl::on_unmap_event(GdkEventAny* event)
{
    Gtk::Window::on_unmap_event(event);

    SetSkipTimeUpdate(true);

    return false;
}

bool MainWndImpl::on_configure_event(GdkEventConfigure* ev)
{
    bool rtn(Gtk::Window::on_configure_event(ev));

    if(!compactMode_ && !browserDisplayMenu_.get_active())
        playlist_.SetTitleWidth(ev->width - 104);

    return rtn;
}

/// @brief Callback function used when quit menu is clicked.
///
/// Stop play thread and call MainManager::Quit.
void MainWndImpl::OnMenuQuit()
{
    mainWndInterface_.Quit();
}

/// @brief Callback function used when compact view menu is clicked.
///
/// When compact mode is activated, hide time display, seek bar, 
/// volumn bar, playlist and resize to small.
void MainWndImpl::ToggleCompactView()
{
    compactMode_ = !compactMode_;

    UpdateDisplay();

    SetDisplayLabel(BoldWeightTitle(displayLabel_.get_text(), compactMode_));
}

void MainWndImpl::OnMenuViewStatusbar()
{
    if(statusbarDisplayMenu_.get_active())
        statusbar_.show();
    else
        statusbar_.hide();
}

/// @brief Callback function used when seek bar is changed.
void MainWndImpl::OnScaleSeek(const double& newvalue)
{
    mainWndInterface_.Seek(static_cast<unsigned long>(newvalue));
}

void MainWndImpl::SetPlayMode()
{
    if(randomPlayMenu_.get_active())
        mainWndInterface_.SetRandomPlay();

    else if(repeatPlayMenu_.get_active())
        mainWndInterface_.SetRepeatPlay();

    else
        mainWndInterface_.SetNormalPlay();
}

void MainWndImpl::SetRepeatCurrent()
{
    mainWndInterface_.SetRepeatCurrentPlay(repeatCurrentPlayMenu_.get_active());
}

void MainWndImpl::UpdateItemRead()
{
    DisplayTitle(playlist_.CurrentItemTitle());
}

void MainWndImpl::UpdatePlay()
{
    DisplayTotalTime(playlist_.CurrentItemTime());
    SetSeekScaleRange(playlist_.CurrentItemTimeSeconds());
    DisplayVisualPanel();
}

void MainWndImpl::UpdateStop()
{
    seekScale_.set_value(0.0);

    ClearVisualPanel();
}

void MainWndImpl::UpdateItemRemoved()
{
    ClearDisplayTitle();
}

/// @brief Callback function used when display label is pressed.
///
/// When display label is double clicked, toggle compace mode.
bool MainWndImpl::OnButtonPressInDisplayLabel(GdkEventButton* ev)
{
    if(ev->type == GDK_2BUTTON_PRESS) {

        ToggleCompactView();

        return true;
    }

    return false;
}

bool MainWndImpl::OnLeftArrowPress(GdkEventButton* ev)
{
    browserDisplayMenu_.set_active(false);

    return true;
}

bool MainWndImpl::OnRightArrowPress(GdkEventButton* ev)
{
    browserDisplayMenu_.set_active(true);

    return true;
}

bool MainWndImpl::OnUpArrowPress(GdkEventButton* ev)
{
    listDisplayMenu_.set_active(false);

    return true;
}

bool MainWndImpl::OnDownArrowPress(GdkEventButton* ev)
{
    listDisplayMenu_.set_active(true);

    return true;
}

/// @brief Update display.
void MainWndImpl::UpdateDisplay()
{
    int currentWidth, currentHeight;
    get_size(currentWidth, currentHeight);

    if(compactMode_) {

        SetBackgroundPixmapInCompactMode();

        SetSkipTimeUpdate(true);

        show_all_children();

        playCtlBox_.hide();
        timeDisplayBox_.hide();
        seekScaleAlignment_.hide();	

        arrowBoxWithVolumn_.hide();

        viewBox_.hide();

        menuBar_.hide();

        statusbar_.hide();

        set_size_request(defaultCompactWidth_, defaultCompactHeight_);
        resize(defaultCompactWidth_, defaultCompactHeight_);
        // instead set_resizable(false)
        SetMaxSizeGeometry(*this, -1, -1);

        prevWidth_ = currentWidth;
        if(currentHeight != defaultFoldHeight_)
            prevHeight_ = currentHeight;

    } else {

        UnsetBackgroundPixmapInCompactMode();

        SetSkipTimeUpdate(false);

        show_all_children();

        compactPlayCtlBox_.hide();

        if(!mainWndInterface_.InPlay())
            ClearVisualPanel();

        if(browserDisplayMenu_.get_active()) {

            rightArrowEventBox_.hide();

            dirPlayWnd_.ShowBrowser();

        } else {

            leftArrowEventBox_.hide();

            dirPlayWnd_.HideBrowser();

        }

        if(currentWidth == defaultCompactWidth_)
            currentWidth = prevWidth_;
        else
            prevWidth_ = currentWidth;

        if(currentHeight == defaultFoldHeight_ ||
           currentHeight == defaultCompactHeight_)
            currentHeight = prevHeight_;
        else
            prevHeight_ = currentHeight;

        if(listDisplayMenu_.get_active()) {

            downArrowEventBox_.hide();

            if(!statusbarDisplayMenu_.get_active())
                statusbar_.hide();

            set_size_request(defaultWidth_, defaultHeight_);
            resize(currentWidth, currentHeight);

            SetMaxSizeGeometry(*this, maxScreenWidth_, maxScreenHeight_);

        } else {

            upArrowEventBox_.hide();

            statusbar_.hide();
            viewBox_.hide();

            set_size_request(defaultWidth_, defaultFoldHeight_);
            resize(currentWidth, defaultFoldHeight_);

            SetMaxSizeGeometry(*this, maxScreenWidth_, -1);
        }
    }
}

/// @brief Set background pixmap when enter compact mode.
void MainWndImpl::SetBackgroundPixmapInCompactMode()
{
    const std::string& bgFile(setting_.GetPreference().GetString("GeneralCompactBgPixmap"));

    if(IsReadable(bgFile)) {

        titleDisplayEventBox_.modify_style(titleDisplayEventBoxDefaultStyle_->copy());
        titleDisplayEventBox_.modify_bg_pixmap(Gtk::STATE_NORMAL, bgFile);
        modify_style(mainWndDefaultStyle_->copy());
        modify_bg_pixmap(Gtk::STATE_NORMAL, bgFile);
    }
}

/// @brief Unset background pixmap when return to normal mode.
inline void MainWndImpl::UnsetBackgroundPixmapInCompactMode()
{
    titleDisplayEventBox_.modify_style(titleDisplayEventBoxDefaultStyle_->copy());
    modify_style(mainWndDefaultStyle_->copy());
}

/// @brief Clear display title.
void MainWndImpl::ClearDisplayTitle()
{
    displayLabel_.set_label("");
    set_title("Sagua");
    mainWndInterface_.TrayIconSetTooltips("Sagua");
}

/// @brief Display string.
/// @param title String to display.
void MainWndImpl::DisplayTitle(const Glib::ustring& title)
{
    SetDisplayLabel(BoldWeightTitle(title, compactMode_));

    set_title("Sagua - " + title);
    mainWndInterface_.TrayIconSetTooltips(title);
}

void MainWndImpl::SyncMixerWidget()
{
    volumnScale_.set_value(mainWndInterface_.GetVolumn());
}

void MainWndImpl::ChangeCompactBgPixmap()
{
    if(compactMode_)
        SetBackgroundPixmapInCompactMode();
}

void MainWndImpl::SetStatusbar(const Glib::ustring& text)
{
    DisconnectTimeoutClearStatusbar();

    statusbar_.pop();
    statusbar_.push(text);
}

void MainWndImpl::SetStatusbarText(const Glib::ustring& text)
{
    statusbarText_ = text;
}

void MainWndImpl::UpdateStatusbar()
{
    DisconnectTimeoutClearStatusbar();

    statusbar_.pop();
    statusbar_.push(statusbarText_);
}

void MainWndImpl::UpdateStatusbarTimeoutClear()
{
    UpdateStatusbar();
    TimeoutClearStatusbar();
}

void MainWndImpl::ClearStatusbar()
{
    statusbar_.pop();
}

void MainWndImpl::TimeoutClearStatusbar()
{
    // need this? anyway double check it.
    DisconnectTimeoutClearStatusbar();

    clearStatusbarTimeout_ = new sigc::connection(Glib::signal_timeout().connect(sigc::mem_fun(*this, &MainWndImpl::TimeoutClearStatusbarSlot), 3000));
}

/// @todo Split errMessage and use set_secondary_text at gtkmm 2.6.x
void MainWndImpl::ShowAudioOpenFailDlg()
{
    const Glib::ustring errMessage(gettext("<span weight=\"bold\" size=\"larger\">Could not open audio device</span>\n\nInvalid audio driver setting or cannot access audio device. Check Audio Driver section in preference, permission on audio device and another program that use audio device.\n"));

    Gtk::MessageDialog errDialog(*this,
                                 errMessage,
                                 true,
                                 Gtk::MESSAGE_ERROR,
                                 Gtk::BUTTONS_CLOSE,
                                 true);

    errDialog.run();
}

void MainWndImpl::Packing(AddItem& addItem)
{
    add(mainBox_);

    Gtk::Menu_Helpers::MenuList* menuList = &fileMenu_.items();
    menuList->push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Add files"), sigc::mem_fun(addItem, &AddItem::FileAdd)) );
    menuList->push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a _directory"), sigc::mem_fun(addItem, &AddItem::DirAdd)) );

#ifdef ENABLE_HTTP
    menuList->push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a _URL"), sigc::mem_fun(addItem, &AddItem::UrlAdd)) );
#endif

    menuList->push_back( Gtk::Menu_Helpers::SeparatorElem());
    menuList->push_back( Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::QUIT, sigc::mem_fun(*this, &MainWndImpl::OnMenuQuit)) );
    menuBar_.items().push_back( Gtk::Menu_Helpers::MenuElem(gettext("_File"), fileMenu_) );

    menuList = &editMenu_.items();
    menuList->push_back( repeatPlayMenu_ );
    menuList->push_back( randomPlayMenu_ );
    menuList->push_back( repeatCurrentPlayMenu_ );
    menuList->push_back( Gtk::Menu_Helpers::SeparatorElem());
    menuList->push_back( Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::PREFERENCES, sigc::mem_fun(mainWndInterface_, &MainWndInterface::RunPrefWnd)) );
    menuBar_.items().push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Edit"), editMenu_) );

    menuList = &viewMenu_.items();
    menuList->push_back( listDisplayMenu_ );
    menuList->push_back( browserDisplayMenu_ );
    menuList->push_back( statusbarDisplayMenu_ );
    menuBar_.items().push_back( Gtk::Menu_Helpers::MenuElem(gettext("_View"), viewMenu_) );

    menuList = &helpMenu_.items();
    menuList->push_back( Gtk::Menu_Helpers::MenuElem(gettext("_About"), sigc::mem_fun(mainWndInterface_, &MainWndInterface::About)) );
    menuBar_.items().push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Help"), helpMenu_) );

    mainBox_.pack_start(menuBar_, Gtk::PACK_SHRINK);

    titleDisplayAlignment_.set_padding(12, 12, 0, 0);
    titleDisplayAlignment_.set_size_request(-1, 52);
    titleDisplayAlignment_.add(displayLabel_);
    titleDisplayEventBox_.add(titleDisplayAlignment_);
    playerBox_.pack_start(titleDisplayEventBox_);

    timeDisplayBox_.pack_start(*manage(new Gtk::Label));
    timeDisplayBox_.pack_start(timeLabel_, Gtk::PACK_SHRINK);
    SetFontSize(timeLabel_, 10);
    timeDisplayBox_.pack_start(totalTimeLabel_, Gtk::PACK_SHRINK);
    SetFontSize(totalTimeLabel_, 10);
    timeDisplayBox_.pack_start(*manage(new Gtk::Label));
    playerBox_.pack_start(timeDisplayBox_);
    seekScale_.set_draw_value(false);
    seekScale_.set_update_policy(Gtk::UPDATE_DELAYED);
    seekScaleAlignment_.add(seekScale_);
    playerBox_.pack_start(seekScaleAlignment_);

    playCtlBox_.pack_start(*manage(new Gtk::Label));
    playCtlBox_.pack_start(prevImgBtn_, Gtk::PACK_SHRINK);
    playCtlBox_.pack_start(stopImgBtn_, Gtk::PACK_SHRINK);
    playCtlBox_.pack_start(pauseImgBtn_, Gtk::PACK_SHRINK);
    playCtlBox_.pack_start(playImgBtn_, Gtk::PACK_SHRINK);
    playCtlBox_.pack_start(nextImgBtn_, Gtk::PACK_SHRINK);
    playCtlBox_.pack_start(*manage(new Gtk::Label));
    playerBox_.pack_start(playCtlBox_);

    compactPlayCtlHBox_.pack_start(*manage(new Gtk::HBox));
    compactPlayCtlHBox_.pack_start(compactPrevImgBtn_, Gtk::PACK_SHRINK);
    compactPlayCtlHBox_.pack_start(compactStopImgBtn_, Gtk::PACK_SHRINK);
    compactPlayCtlHBox_.pack_start(compactPauseImgBtn_, Gtk::PACK_SHRINK);
    compactPlayCtlHBox_.pack_start(compactPlayImgBtn_, Gtk::PACK_SHRINK);
    compactPlayCtlHBox_.pack_start(compactNextImgBtn_, Gtk::PACK_SHRINK);
    compactPlayCtlHBox_.pack_start(*manage(new Gtk::HBox));
    compactPlayCtlBox_.set_spacing(12);
    compactPlayCtlBox_.pack_start(*manage(new Gtk::HBox));
    compactPlayCtlBox_.pack_start(compactPlayCtlHBox_);
    playerBox_.pack_start(compactPlayCtlBox_);

    playerBoxWithVolumn_.pack_start(playerBox_);
    volumnScale_.set_inverted(true);
    volumnScale_.set_draw_value(false);
    volumnScale_.set_value(mainWndInterface_.GetVolumn());
    volumnAlignment_.add(volumnScale_);

    leftArrowEventBox_.add(leftArrow_);
    leftArrowEventBox_.set_events(Gdk::BUTTON_PRESS_MASK);
    arrowBox_.pack_start(leftArrowEventBox_, Gtk::PACK_SHRINK);

    rightArrowEventBox_.add(rightArrow_);
    rightArrowEventBox_.set_events(Gdk::BUTTON_PRESS_MASK);
    arrowBox_.pack_start(rightArrowEventBox_, Gtk::PACK_SHRINK);

    upArrowEventBox_.add(upArrow_);
    upArrowEventBox_.set_events(Gdk::BUTTON_PRESS_MASK);
    arrowBox_.pack_start(upArrowEventBox_, Gtk::PACK_SHRINK);

    downArrowEventBox_.add(downArrow_);
    downArrowEventBox_.set_events(Gdk::BUTTON_PRESS_MASK);
    arrowBox_.pack_start(downArrowEventBox_, Gtk::PACK_SHRINK);

    arrowBoxWithVolumn_.pack_start(volumnAlignment_);
    arrowBoxWithVolumn_.pack_start(arrowBox_, Gtk::PACK_SHRINK);
    playerBoxWithVolumn_.pack_start(arrowBoxWithVolumn_, Gtk::PACK_SHRINK);

    mainBox_.pack_start(playerBoxWithVolumn_, Gtk::PACK_SHRINK);

    mainBox_.pack_start(viewBox_);
    mainBox_.pack_start(statusbar_, Gtk::PACK_SHRINK);
}

void MainWndImpl::InitialDisplay()
{
    LastState& lastState(setting_.GetLastState());

    repeatPlayMenu_.set_active(lastState.GetPlayRepeat());
    randomPlayMenu_.set_active(lastState.GetPlayRandom());
    repeatCurrentPlayMenu_.set_active(lastState.GetPlayRepeatCurrent());

    if(lastState.GetWindowWidth() > defaultWidth_)
        prevWidth_ = lastState.GetWindowWidth();

    if(lastState.GetWindowHeight() > defaultHeight_)
        prevHeight_ = lastState.GetWindowHeight();

    resize(prevWidth_, prevHeight_);

    if(lastState.GetViewStatusBar())
        statusbar_.show();

    statusbarDisplayMenu_.set_active(lastState.GetViewStatusBar());

    listDisplayMenu_.set_active(lastState.GetViewList());
    browserDisplayMenu_.set_active(lastState.GetViewBrowser());

    viewBox_.pack_start(dirPlayWnd_.GetWidget());

    UpdateDisplay();

    ClearDisplayTitle();
}

/// @brief Save last state.
void MainWndImpl::SaveState()
{
    LastState& lastState(setting_.GetLastState());

    int lastWidth = 0;
    int lastHeight = 0;
    if(prevWidth_ != defaultWidth_ || prevHeight_ != defaultHeight_) {

        lastWidth = prevWidth_;
        lastHeight = prevHeight_;
    } 

    if(!compactMode_ && listDisplayMenu_.get_active())
        get_size(lastWidth, lastHeight);

    lastState.SetWindowWidth(lastWidth);
    lastState.SetWindowHeight(lastHeight);
    lastState.SetPlayRepeat(repeatPlayMenu_.get_active());
    lastState.SetPlayRepeatCurrent(repeatCurrentPlayMenu_.get_active());
    lastState.SetPlayRandom(randomPlayMenu_.get_active());
    lastState.SetViewBrowser(browserDisplayMenu_.get_active());
    lastState.SetViewList(listDisplayMenu_.get_active());
    lastState.SetViewCompact(compactMode_);
    lastState.SetViewStatusBar(statusbarDisplayMenu_.get_active());
    lastState.SetLastBrowseDirectory(dirPlayWnd_.GetSelectedDirectory());

    lastState.Save();
}

void MainWndImpl::ConnectSignals()
{
    repeatPlayMenu_.signal_toggled().connect(sigc::mem_fun(*this, &MainWndImpl::SetPlayMode) );
    randomPlayMenu_.signal_toggled().connect(sigc::mem_fun(*this, &MainWndImpl::SetPlayMode) );
    repeatCurrentPlayMenu_.signal_toggled().connect(sigc::mem_fun(*this, &MainWndImpl::SetRepeatCurrent) );
    listDisplayMenu_.signal_toggled().connect(sigc::mem_fun(*this, &MainWndImpl::UpdateDisplay));
    browserDisplayMenu_.signal_toggled().connect(sigc::mem_fun(*this, &MainWndImpl::UpdateDisplay));
    statusbarDisplayMenu_.signal_toggled().connect(sigc::mem_fun(*this, &MainWndImpl::OnMenuViewStatusbar));

    titleDisplayEventBox_.signal_button_press_event().connect(sigc::mem_fun(*this, &MainWndImpl::OnButtonPressInDisplayLabel) );

    seekScale_.signal_adjust_bounds().connect(sigc::mem_fun(*this, &MainWndImpl::OnScaleSeek) );

    prevImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::Prev) );
    stopImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::Stop) );
    pauseImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::PauseToggle) );
    playImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::PlayStart) );
    nextImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::Next) );
    compactPrevImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::Prev) );
    compactStopImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::Stop) );
    compactPauseImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::PauseToggle) );
    compactPlayImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::PlayStart) );
    compactNextImgBtn_.SignalPress().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::Next) );

    volumnScale_.signal_adjust_bounds().connect(sigc::mem_fun(mainWndInterface_, &MainWndInterface::SetVolumn) );

    leftArrowEventBox_.signal_button_press_event().connect(sigc::mem_fun(*this, &MainWndImpl::OnLeftArrowPress) );
    rightArrowEventBox_.signal_button_press_event().connect(sigc::mem_fun(*this, &MainWndImpl::OnRightArrowPress) );
    upArrowEventBox_.signal_button_press_event().connect(sigc::mem_fun(*this, &MainWndImpl::OnUpArrowPress) );
    downArrowEventBox_.signal_button_press_event().connect(sigc::mem_fun(*this, &MainWndImpl::OnDownArrowPress) );
}
