/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file misc.cc
/// @brief Miscellaneous functions.

#include <cstring>
#include <fstream>

#include "../include/misc.h"

/// @brief Get extention part of filename.
std::string GetExt(const char* filename)
{
    filename = std::strrchr(filename, '.');

    if(0 != filename) {

        ++filename;

        char ext[4];
        unsigned int index = 0;
        while(index < 3 && filename[index] != '\0') {

            ext[index] = std::tolower(filename[index]);

            ++index;
        }

        ext[index] = '\0';

        return ext;

    } else
        return std::string();
}

Glib::ustring ConvertSecondToTimeString(const unsigned long& second)
{
    char timeStr[18];

    if(second < 3600)
        std::snprintf(timeStr, sizeof(timeStr), "%02lu:%02lu", second / 60, second % 60);
    else
        std::snprintf(timeStr, sizeof(timeStr), "%02lu:%02lu:%02lu", second / 3600, second % 3600 / 60, second % 3600 % 60);
    
    return timeStr;
}

bool HasHttpProtocolString(const std::string& filePath)
{
    if(filePath.size() < 7)
        return false;

    // convert to lower case and strcmp ?

    if((filePath[0] == 'h' || filePath[0] == 'H') &&
       (filePath[1] == 't' || filePath[1] == 'T') &&
       (filePath[2] == 't' || filePath[2] == 'T') &&
       (filePath[3] == 'p' || filePath[3] == 'P') &&
       filePath[4] == ':' && filePath[5] == '/' &&
       filePath[6] == '/')
        return true;
    else
        return false;
}

bool HasFileProtocolString(const std::string& str)
{
    if(str.size() < 7)
        return false;

    // convert to lower case and strcmp ?

    if((str[0] == 'f' || str[0] == 'F') &&
       (str[1] == 'i' || str[1] == 'I') &&
       (str[2] == 'l' || str[2] == 'L') &&
       (str[3] == 'e' || str[3] == 'E') &&
       str[4] == ':' && str[5] == '/' &&
       str[6] == '/')
        return true;
    else
        return false;
}

/// @brief Read config file and put to map.
/// @param filename Config filename.
/// @param configMap Destination map.
/// @param convertLowerCase Convert key and value to lower case.
/// @param valueBreakAtSpace Read value to space character. 
/// @return Success of open config file.
bool ConfigfileToMap(const std::string& filename, std::map<std::string, std::string>& configMap, const bool& convertLowerCase, const bool& valueBreakAtSpace)
{
    std::ifstream prefFile(filename.c_str());

    if(prefFile) {

        typedef std::string::iterator iter;

        std::string eachLine;

        while(std::getline(prefFile, eachLine)) {

            if(IsComment(eachLine))
                continue;

            const iter end(eachLine.end());

            // Convert string to lower case
            if(convertLowerCase)
                StringToLower(eachLine);

            // Eleminate front space character 
            iter i = find_if(eachLine.begin(), end, std::not1(IsCharSpace()));

            if(i != end) {

                // Find '=' character
                iter j = find_if(i, end, IsCharEqual());

                if(i != j) {
                    // Eleminate space character between i to j and that is key to insert.

                    iter keyBegin(i);
                    iter keyEnd(find_if(i, j, IsCharSpace()));

                    // Find value to insert.
                    i = find_if(++j, end, std::not1(IsCharSpace()));

                    if(i != end) {

                        if(valueBreakAtSpace)
                            j = find_if(i, end, IsCharSpace());
                        else
                            j = end;

                        configMap[std::string(keyBegin, keyEnd)] = std::string(i, j);
                    }
                }
            }
        }

        return true;
    }

    return false;
}

/// @brief Convert string to UTF8 without fail string.
/// @param orig Original string to convert.
/// @param charset From charset.
/// @return Converted string.
Glib::ustring StringToUTF8::StringToUTF8WithoutFail(const Glib::ustring& orig, const std::string& charset) const
{
    try {

        return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch(Glib::ConvertError& err) {

        return Glib::ustring();
    }
}

Glib::ustring StringToUTF8::StringToUTF8WithoutFail(const std::string& orig, const std::string& charset) const
{
    try {

        return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch(Glib::ConvertError& err) {

        return Glib::ustring();
    }
}

Glib::ustring StringToUTF8::StringToUTF8WithoutFail(const char* orig, const std::string& charset) const
{
    try {

        return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch(Glib::ConvertError& err) {

        return Glib::ustring();
    }
}

/// @brief Convert string to UTF8 with fail string.
/// @param orig Original string to convert.
/// @param fail Retry string when convert is failed.
/// @param charset From charset.
/// @return Converted string.
/// @sa Glib::ustring StringToUTF8
Glib::ustring StringToUTF8::StringToUTF8WithFail(const Glib::ustring& orig, const std::string& fail, const std::string& charset) const
{
    try {

        return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch(Glib::ConvertError& err) {

        return StringToUTF8WithoutFail(fail, charset);
    }
}

Glib::ustring StringToUTF8::StringToUTF8WithFail(const std::string& orig, const std::string& fail, const std::string& charset) const
{
    try {

        return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch(Glib::ConvertError& err) {

        return StringToUTF8WithoutFail(fail, charset);
    }
}

Glib::ustring StringToUTF8::StringToUTF8WithFail(const char* orig, const std::string& fail, const std::string& charset) const
{
    try {

        return Glib::convert_with_fallback(orig, "UTF-8", charset);

    } catch(Glib::ConvertError& err) {

        return StringToUTF8WithoutFail(fail, charset);
    }
}
