/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_ogg.cc
/// @brief Class for play ogg source file.

#include "../include/player_ogg.h"
#include "../include/mainmanager.h"
#include "../include/trace.h"

/// @brief Constructor of OGGPlayer.
///
/// Connect dispatchers to functions.
template <class Buffer>
VORBIS::OGGPlayer<Buffer>::OGGPlayer(Buffer& buffer, const std::string& filename)
    : ovFile_(filename.c_str()),
      decodeState_(STOP),
      buflen_(4096),
      buffer_(buffer),
      playerInterface_(MainManagerImpl::Instance())
{
    signalFinish_.connect(sigc::mem_fun(playerInterface_, &PlayerInterface::Finished));
}

/// @brief Start player loop.
///
/// Decode file and send decoded samples to play function. Timer update and
/// audio open, close is also in this function.
/// @todo Avoid infinite loop problem on EOF and OV_HOLE in sometimes.
template <class Buffer>
void VORBIS::OGGPlayer<Buffer>::DoStart()
{
    DEBUGTRACE("VORBIS::OGGPlayer<Buffer>::Start");

    decodeState_ = PLAY;

    float** buf = 0;
    int ret = 0;
    int bs = 0;
    for(;;) {	

        if(decodeState_ == STOP) {

            DEBUGTRACE_MSG("decoedState_ == STOP");

            break;
        }
	
        {
            Glib::Mutex::Lock lock(vfMutex_);	
	    
            ret = ovFile_.Read(&buf, buflen_, &bs);
        }


        if(0 == ret) {

            DEBUGTRACE_MSG("End of file");

            break;
	    
        } else if(ret > 0) {

            if(bs != 0) {

                DEBUGTRACE_MSG("bs != 0");

                break;
            }

            if(!playerInterface_.CheckOpenDevice()) {

                buffer_.Close(true);

                DEBUGTRACE_MSG("OpenDevice");

                if(!playerInterface_.OpenDevice(16, ovFile_.GetChannels(), ovFile_.GetSampleRate()))
                    break;
            }

            buffer_.Push(buf, ret);
        } 
    }
    
    buffer_.Close(decodeState_ == PLAY);
		
    signalFinish_();
}

/// @brief Set decodeState to STOP.
///
/// Set decodeState to STOP. Detect STOP in Start() function.
template <class Buffer>
void VORBIS::OGGPlayer<Buffer>::DoStop()
{
    decodeState_ = STOP;
}

/// @brief Seek to wanted time.
/// @param seconds wanted time to seek in seconds.
template <class Buffer>
void VORBIS::OGGPlayer<Buffer>::DoSeek(unsigned long time)
{
    Glib::Mutex::Lock lock(vfMutex_);

    ovFile_.Seek(time);

    buffer_.SetCurrentTime(time);
}

/// @brief Create OGGPlayer object.
///
/// Used in PlayerFactory.
template <class Buffer>
std::auto_ptr<Player> CreateOGGPlayer(Buffer& buffer, const std::string& filename)
{
    return std::auto_ptr<Player>(new VORBIS::OGGPlayer<Buffer>(buffer, filename));
}
