/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// wma decoder came from bmp-wma plugin.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_wma.cc
/// @brief Class for play wma source file.

#include "../include/player_wma.h"
#include "../include/mainmanager.h"
#include "../include/exception.h"

/// @brief Constructor of WMAPlayer.
///
/// Connect dispatchers to functions.
template <class Buffer>
FFMPEG::WMAPlayer<Buffer>::WMAPlayer(Buffer& buffer, const std::string& filename)
    : decodeState_(STOP),
      seekSeconds_(0),
      bufSize_(4096),
      count_(bufSize_/2),
      index_(0),
      buffer_(buffer),
      playerInterface_(MainManagerImpl::Instance()),
      avCodec_(0),
      avCodecContext_(0),
      avFormatContext_(0)
{
    avcodec_init();
    avcodec_register_all();
    av_register_all();

    if(av_open_input_file(&avFormatContext_, filename.c_str(), 0, 0, 0) < 0)
        throw Exception::WmaException("av_open_input_file failed");

    for(index_ = 0; index_ < avFormatContext_->nb_streams; ++index_) {

        avCodecContext_ = &avFormatContext_->streams[index_]->codec;
        if(avCodecContext_->codec_type == CODEC_TYPE_AUDIO)
            break;	
    }

    av_find_stream_info(avFormatContext_);

    try {

        AVCodecSelect();

    } catch(const Exception::SaguaException& ex) {

        av_close_input_file(avFormatContext_);

        throw;
    }

    signalFinish_.connect(sigc::mem_fun(playerInterface_, &PlayerInterface::Finished));
}

/// @brief Start player loop.
///
/// Decode file and send decoed samples to play function. Timer update and
/// audio open, close is also in this function.
template <class Buffer>
void FFMPEG::WMAPlayer<Buffer>::DoStart()
{
    unsigned int channels = avCodecContext_->channels;
    if(!playerInterface_.CheckOpenDevice()) {

        buffer_.Close(true);
        if(!playerInterface_.OpenDevice(16, channels, avCodecContext_->sample_rate)) {

            signalFinish_();

            return;
        }
    }

    decodeState_ = PLAY;

    int size = 0;
    int len = 0;
    int out_size = 0;
    uint8_t* inbuf_ptr = 0;

    short int* outbuf = (short int*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);

    for(;;) {

        {
            Glib::Mutex::Lock lock(avMutex_);

            if(decodeState_ == STOP)
                break;
            else if(decodeState_ == SEEK) {
                av_seek_frame(avFormatContext_, index_, seekSeconds_ * 1000000LL);
                buffer_.SetCurrentTime(seekSeconds_);
                seekSeconds_ = 0;
                decodeState_ = PLAY;
            }
        }

        if(av_read_frame(avFormatContext_, &avPacket_) < 0)
            break;

        size = avPacket_.size;
        inbuf_ptr = avPacket_.data;

        if(size == 0)
            break;

        while(size > 0) {
            len = avcodec_decode_audio(avCodecContext_, 
                                       outbuf,
                                       &out_size,
                                       inbuf_ptr,
                                       size);

            if(len < 0) 
                break;

            if(out_size <= 0)
                continue;
            else {

                unsigned int offset = 0;
                unsigned int outbufSize = out_size / 2;
                while(outbufSize >= bufSize_) {

                    buffer_.Push(outbuf + offset, count_);

                    offset += bufSize_;
                    outbufSize -= bufSize_;
                }

            }

            size -= len;
            inbuf_ptr += len;

            if(avPacket_.data)
                av_free_packet(&avPacket_);	
        }
    }

    free(outbuf);

    Clean();

    buffer_.Close(decodeState_ == PLAY);

    signalFinish_();
}

/// @brief Set decodeState to STOP.
///
/// Set decodeState to STOP. Detect STOP in Statr() function.
template <class Buffer>
void FFMPEG::WMAPlayer<Buffer>::DoStop()
{
    Glib::Mutex::Lock lock(avMutex_);

    decodeState_ = STOP;
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds..
///
/// Set decodeState to SEEK. Detect SEEK in Start() function.
/// @param seconds wanted time to seek in seconds.
template <class Buffer>
void FFMPEG::WMAPlayer<Buffer>::DoSeek(unsigned long time)
{
    Glib::Mutex::Lock lock(avMutex_);

    decodeState_ = SEEK;

    seekSeconds_ = time;

    buffer_.SetCurrentTime(time);
}

template <class Buffer>
void FFMPEG::WMAPlayer<Buffer>::AVCodecSelect()
{
    avCodec_ = avcodec_find_decoder(avCodecContext_->codec_id);
    if(!avCodec_)
        throw Exception::WmaException("avcodec_find_decoder failed");

    if(avcodec_open(avCodecContext_, avCodec_) < 0)
        throw Exception::WmaException("avcodec_open failed");
}

/// @brief Create WMAPlayer object.
///
/// Used in PlayerFactory.
template <class Buffer>
std::auto_ptr<Player> CreateWMAPlayer(Buffer& buffer, const std::string& filename)
{
    return std::auto_ptr<Player>(new FFMPEG::WMAPlayer<Buffer>(buffer, filename));
}
