/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_aac.cc
/// @brief Class for mpeg 4 tag processing source file.

#include <string>

#include "../include/tag_aac.h"

AACTag::AACTag(const std::string& filename)
{
    FileHandler file(filename.c_str(), "rb");

    if(FAAD::IsMP4File(file.Get()))
        ProcessMP4(file.Get());
    else
        ProcessAAC(file.Get());

    if(title_.empty())
        title_ = StringToUTF8::Instance().Convert(Glib::path_get_basename(filename));

    totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);
}

/// @brief Get information from AAC type file.
///
/// @todo Need testing.
void AACTag::ProcessAAC(FILE* file)
{
    FAAD::AACBuffer aacBuffer(file);

    if(!aacBuffer.InitBuffer(FAAD_MIN_STREAMSIZE))
        return;

    totalTimeSeconds_ = static_cast<unsigned long>(aacBuffer.GetLength());
}

void AACTag::ProcessMP4(FILE* file)
{
    try {

        MP4 mp4(file);

        mp4.GetTotalTimeSeconds(totalTimeSeconds_);
        mp4.GetTitle(title_);
        mp4.GetArtist(artist_);
        mp4.GetAlbum(album_);
        mp4.GetTrack(track_);
        mp4.GetDate(year_);
        mp4.GetGenre(genre_);
        mp4.GetComment(comment_);

    } catch(const Exception::SaguaException& ex) {

    }
}

void AACTag::ProcessYear()
{}

void AACTag::ProcessTrack()
{}

void AACTag::ProcessAlbum()
{}

void AACTag::ProcessGenre()
{}

void AACTag::ProcessArtist()
{}

void AACTag::ProcessComment()
{}

/// @brief Create AACTag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateAACTag(const std::string& filename)
{
    return std::auto_ptr<Tag>(new AACTag(filename));
}

/// @throw FaadException Exception on init.
MP4::MP4(FILE* file)
    : track_(-1),
      frameSize_(1024),
      decoder_(faacDecOpen()),
      mp4ffFile_(0),
      mp4Cb_(FAAD::SetMP4ffCallback(file))
{
    try {

        Init();

    } catch(const std::exception& ex) {

        faacDecClose(decoder_);

        throw;
    }

    mp4ffMetaCallbacks_["title"] = mp4ff_meta_get_title;
    mp4ffMetaCallbacks_["artist"] = mp4ff_meta_get_artist;
    mp4ffMetaCallbacks_["album"] = mp4ff_meta_get_album;
    mp4ffMetaCallbacks_["date"] = mp4ff_meta_get_date;
    mp4ffMetaCallbacks_["genre"] = mp4ff_meta_get_genre;
    mp4ffMetaCallbacks_["comment"] = mp4ff_meta_get_comment;
}

/// @throw FaadException Exception on init.
void MP4::Init()
{
    mp4ffFile_ = mp4ff_open_read(mp4Cb_.get());
    if(0 == mp4ffFile_)
        throw Exception::FaadException("mp4ff_open_read failed");

    track_ = FAAD::GetAACTrack(mp4ffFile_);
    if(track_ < 0) {

        mp4ff_close(mp4ffFile_);

        throw Exception::FaadException("Cannot get aac track");
    }

    unsigned char* buffer = 0;
    unsigned int bufferSize = 0;
    mp4ff_get_decoder_config(mp4ffFile_, track_, &buffer, &bufferSize);

    unsigned long sampleRate = 0;
    unsigned char channels = 0;
    if(faacDecInit2(decoder_, buffer, bufferSize, &sampleRate, &channels) < 0) {
        mp4ff_close(mp4ffFile_);

        throw Exception::FaadException("Cannot init faac decoder");
    }

    if(0 != buffer) {

        if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC_) >= 0) {
            if(mp4ASC_.frameLengthFlag == 1)
                frameSize_ = 960;
            if(mp4ASC_.sbr_present_flag == 1)
                frameSize_ *= 2;	
        }

        free(buffer);
        buffer = 0;
        bufferSize = 0;
    }
}

inline void MP4::GetTotalTimeSeconds(unsigned long& totalTimeSeconds)
{
    totalTimeSeconds = static_cast<unsigned long>(mp4ff_num_samples(mp4ffFile_, track_) * (frameSize_ - 1.0) / mp4ASC_.samplingFrequency);
}

void MP4::GetMetaData(const char* const key, Glib::ustring& output)
{
    Mp4ffMetaCallbacks::const_iterator iter = mp4ffMetaCallbacks_.find(key);
    if(iter != mp4ffMetaCallbacks_.end()) {

        char* value = 0;
        if(iter->second(mp4ffFile_, &value)) {

            if(0 != value) {

                output = StringToUTF8::Instance().Convert(value);

                free(value);

            }
        }
    }
}
