/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file trayicon.cc
/// @brief Class for trayicon source file

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <libintl.h>

#include <gtkmm/stock.h>

#include "../include/trayicon.h"
#include "../include/command.h"

TrayIcon::TrayIcon(const char* imageFile, const char* id)
    : trayIconImage_(imageFile),
      eggTrayIcon_(egg_tray_icon_new(id))
{
    EventBoxInit();

    // If you use Glib::wrap to wrap to gtkmm object, you'll get warning about
    // Gtk::manage to Gtk::Window.
    // It is just warning, but i don't want to see warning. So i'll convert to
    // Gtk+ object.
    gtk_container_add(GTK_CONTAINER(eggTrayIcon_), trayIconEventBox_.Gtk::Widget::gobj());
    
    gtk_widget_show_all(GTK_WIDGET(eggTrayIcon_));
}

TrayIcon::~TrayIcon()
{}

void TrayIcon::EventBoxInit()
{
    trayIconEventBox_.add_events(Gdk::BUTTON_PRESS_MASK | Gdk::SCROLL_MASK);
    trayIconEventBox_.signal_button_press_event().connect(sigc::mem_fun(*this, &TrayIcon::OnBtnPress) );

    trayIconEventBox_.signal_scroll_event().connect(sigc::mem_fun(*this, &TrayIcon::OnScroll) );
    trayIconEventBox_.add(trayIconImage_);
}

/// @brief Callback function for TrayIcon EventBox's scroll signal.
///
/// When mouse wheeling, adjust volumn.
/// @param ev GdkEvent
bool TrayIcon::OnScroll(GdkEventScroll* ev)
{
    if(ev->direction == GDK_SCROLL_UP)
        signalWheelUp_();

    if(ev->direction == GDK_SCROLL_DOWN)
        signalWheelDown_();

    return false;
}

/// @brief Callback function for TrayIcon EventBox's button press signal.
///
/// If button is first, act setting in preference.
/// If button is third, pop up trayicon menu.
/// @param ev GdkEvent
bool TrayIcon::OnBtnPress(GdkEventButton* ev)
{

    if(ev->type == GDK_BUTTON_PRESS) {

        if(ev->button ==1)
            signalFirstBtnPress_();
	
        if(ev->button == 3) { 
            trayIconMenu_.popup(sigc::mem_fun(*this, &TrayIcon::MenuPositionCalc), ev->button, ev->time);
        }

    }
    
    return false;
}

void TrayIcon::GetWindowOrigin(int&x, int& y)
{
    Glib::RefPtr<Gdk::Window> gdkWindow(trayIconEventBox_.get_window());
    gdkWindow->get_origin(x, y);
}

int TrayIcon::GetWindowHeight()
{
    int width, height;

    Glib::RefPtr<Gdk::Window> gdkWindow(trayIconEventBox_.get_window());
    gdkWindow->get_size(width, height);

    return height;
}

void TrayIcon::GetMenuSize(int& width, int& height)
{
    if(!trayIconMenu_.is_realized())
        trayIconMenu_.Realize();

    trayIconMenu_.get_window()->get_size(width, height);
}

int TrayIcon::GetMenuWidth()
{
    int menuWidth, menuHeight;

    GetMenuSize(menuWidth, menuHeight);

    return menuWidth;
}

int TrayIcon::GetMenuHeight()
{
    int menuWidth, menuHeight;

    GetMenuSize(menuWidth, menuHeight);

    return menuHeight;
}

int TrayIcon::GetTrayIconWidth()
{
    Glib::RefPtr<Gdk::Screen> gdkScreen(trayIconEventBox_.get_screen());

    return gdkScreen->get_width();
}

int TrayIcon::GetTrayIconHeight()
{
    Glib::RefPtr<Gdk::Screen> gdkScreen(trayIconEventBox_.get_screen());

    return gdkScreen->get_height();
}

/// @brief Menu position calculation when menu pop up.
/// @param x
/// @param y
/// @param pushIn
void TrayIcon::MenuPositionCalc(int& x, int& y, bool& pushIn)
{
    GetWindowOrigin(x, y);

    if(y + GetWindowHeight() < GetTrayIconHeight() / 2)
        y += GetWindowHeight();
    else
        y -= GetMenuHeight();

    if(x > GetTrayIconWidth() - GetMenuWidth())
        x = GetTrayIconWidth() - GetMenuWidth();
}

SaguaTrayIcon::SaguaTrayIcon(Command& command)
    : TrayIcon(PIXMAPSDIR "sagua-small.png", gettext("sagua")),
      mainWndOpenMenu_(gettext("_Open Window"), true)
{
    ConnectSignals(command);

    InitMenu(command);

    mainWndOpenMenu_.show();
}

void SaguaTrayIcon::ConnectSignals(Command& command)
{
    signalFirstBtnPress_.connect(command.GetTrayIconAction());
    signalWheelUp_.connect(command.GetVolumnUp());
    signalWheelDown_.connect(command.GetVolumnDown());
}

void SaguaTrayIcon::InitMenu(Command& command)
{
    mainWndOpenMenu_.signal_activate().connect(command.GetOpenMainWnd());

    Gtk::Menu_Helpers::MenuList& menuList = trayIconMenu_.items();
    menuList.push_back( mainWndOpenMenu_ );
    menuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("P_rev"), command.GetPlayCtrlPrev()) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Play"), command.GetPlayCtrlPlay()) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Pa_use"), command.GetPlayCtrlPause()) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Stop"), command.GetPlayCtrlStop()) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Next"), command.GetPlayCtrlNext()) );
    menuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_File information"), command.GetFileInfo()) );
    menuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_About"), command.GetAbout()) );
    menuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    menuList.push_back( Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::QUIT, command.GetQuit()) );
}
